################################################################################
#
#  CDDL HEADER START
#
#  The contents of this file are subject to the terms of the Common Development
#  and Distribution License Version 1.0 (the "License").
#
#  You can obtain a copy of the license at
#  http:# www.opensource.org/licenses/CDDL-1.0.  See the License for the
#  specific language governing permissions and limitations under the License.
#
#  When distributing Covered Code, include this CDDL HEADER in each file and
#  include the License file in a prominent location with the name LICENSE.CDDL.
#  If applicable, add the following below this CDDL HEADER, with the fields
#  enclosed by brackets "[]" replaced with your own identifying information:
#
#  Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
#
#  CDDL HEADER END
#
#  Copyright (c) 2017-2019, Regents of the University of Minnesota.
#  All rights reserved.
#
#  Contributor(s):
#     Ellad B. Tadmor
#
################################################################################
"""
Helper routines for KIM Verification Checks

"""

import math
import os
import sys
import textwrap
import time

import jinja2
import numpy as np
from ase import Atoms

__all__ = [
    "KIMVCError",
    "VerificationCheck",
    "setup_and_run_vc",
    "vc_stripall",
    "vc_letter_grade_machine_precision",
]


################################################################################
class KIMVCError(Exception):
    def __init__(self, msg):
        # Call the base class constructor with the parameters it needs
        super(KIMVCError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg


################################################################################
def vc_stripall(string):
    """
    In "string", remove leading and trailing newlines, replace internal newlines
    with spaces, and remove carriage returns.

    Primary purpose is to process docstring at top of VC code and convert to
    vc_description.
    """
    return string.strip().replace("\n", " ").replace("\r", "")


################################################################################
def vc_letter_grade_machine_precision(error, Amax=7.5, Bmax=10.0, Cmax=12.5, Dmax=15.0):
    """
    This function assigns a letter grade in the range A-D and F for fail
    based on the obtained 'error' as compared with the machine precision
    of the computation.
    """
    if not Amax < Bmax < Cmax < Dmax:
        raise KIMVCError("ERROR: This condition not satisfied: Amax<Bmax<Cmax<Dmax")
    eps = np.finfo(float).eps
    if abs(error) < eps:
        letter_grade = "A"
    else:
        score = math.log10(abs(error) / eps)
        if score <= Amax:
            letter_grade = "A"
        elif score <= Bmax:
            letter_grade = "B"
        elif score <= Cmax:
            letter_grade = "C"
        elif score <= Dmax:
            letter_grade = "D"
        else:
            letter_grade = "F"
    # Return with assigned letter grade
    comment = (
        "The letter grade {0:} was assigned because the normalized error in the "
        "computation was {1:.5e} compared with a machine precision of {2:.5e}. The "
        "letter grade was based on 'score=log10(error/eps)', with ranges "
        "A=[0, {3:.1f}], B=({3:.1f}, {4:.1f}], C=({4:.1f}, {5:.1f}], D=({5:.1f}, "
        "{6:.1f}), F>{6:.1f}.  'A' is the best grade, and 'F' indicates failure."
    ).format(letter_grade, error, eps, Amax, Bmax, Cmax, Dmax)
    return letter_grade, comment


################################################################################
def setup_and_run_vc(
    do_vc,
    model,
    vc_name,
    vc_author,
    vc_description,
    vc_category,
    vc_grade_basis,
    vc_files,
    vc_debug,
):
    """
    Construct VerificationCheck object containing all info related to VC,
    and perform the VC by calling the user supplied `do_vc` function.  The VC
    report and property instance re generated.  If `vc_debug` is True, errors
    generate exceptions will full traceback, otherwise exceptions are trapped
    and an error message is reported.
    """

    # Define VC object and do verification check
    vc = VerificationCheck(
        vc_name,
        vc_author,
        vc_description,
        vc_category,
        vc_grade_basis,
        vc_files,
        vc_debug,
    )
    try:
        with vc:
            # Perform verification check and get grade
            try:
                vc_grade, vc_comment = do_vc(model, vc)
                vc.rwrite("Grade: {}".format(vc_grade))
                vc.rwrite("")
                vc.rwrite("Comment: " + vc_comment)
                vc.rwrite("")
            except BaseException as e:
                vc_grade = "N/A"
                vc_comment = "Unable to perform verification check due to an error."
                if vc_debug:
                    raise
                else:
                    vc.rwrite(
                        "\nERROR: Unable to perform verification check.\n\n Message = "
                        + str(e)
                    )
            finally:
                # Pack results in a dictionary and write VC property instance
                results = {
                    "vc_name": vc_name,
                    "vc_description": vc_description,
                    "vc_category": vc_category,
                    "vc_grade_basis": vc_grade_basis,
                    "vc_grade": vc_grade,
                    "vc_comment": vc_comment,
                    "vc_files": vc_files,
                }
                vc.write_results(results)
    except Exception as e:
        if vc_debug:
            raise
        else:
            sys.stderr.write(
                "\nERROR: Unable to initalize verification check.\n\n Message = "
                + str(e)
            )


################################################################################
class VerificationCheck(object):
    """
    Features routines related to outputting VC information and results.
    """

    ############################################################################
    def rwrite(self, string):
        """
        Write 'string' with appended newline to report and echo to stdout
        """
        if not self.report_ready:
            raise KIMVCError(
                "ERROR: Trying to write out string to "
                "report file that is not ready.\n\n"
                '       string =  "{}"\n'.format(string)
            )

        self.report.write(string + "\n")
        sys.stdout.write(string + "\n")

    ############################################################################
    def __init__(
        self,
        vc_name,
        vc_author,
        vc_description,
        vc_category,
        vc_grade_basis,
        vc_files,
        vc_debug,
    ):
        """
        Initialize a VC object.
        """
        # Initialize object variables
        self.output_ready = False  # Output directory not setup
        self.report_ready = False  # Report file not open for writing
        self.vc_name = vc_name
        self.vc_description = vc_description
        self.vc_author = vc_author
        self.vc_category = vc_category
        self.vc_grade_basis = vc_grade_basis
        self.vc_files = vc_files
        self.vc_debug = vc_debug

    ############################################################################
    def __enter__(self):
        """
        Beginning with block:
        Send message to stdout, open report file, print header,
        and return report file object.
        """
        # Send message to stdout that VC calculation is starting
        print()
        print(
            "=== Verification check {} start ({}) ==="
            "".format(self.vc_name, time.strftime("%Y-%m-%d %H:%M:%S"))
        )
        # Create directory 'output' if it does not already exist:
        try:
            if not os.path.exists("output"):
                os.makedirs("output")
            self.output_ready = True
        except (OSError, IOError):
            raise KIMVCError("ERROR: Unable to create directory 'output'.\n")
        # Open report file and write header
        try:
            self.report = open(
                os.path.abspath(os.path.join("output", "report.txt")),
                "w",
                encoding="utf-8",
            )
            self.report_ready = True
        except (OSError, IOError):
            raise KIMVCError(
                "ERROR: Unable to open verification check report file for writing.\n"
            )
        # Output header to report file
        padside = 5
        c = "!"
        content_line = "VERIFICATION CHECK: " + self.vc_name
        content_width = len(content_line)
        content_line += " " * (content_width - len(content_line))
        width = 2 * padside + 4 + content_width
        self.rwrite(c * width)
        self.rwrite(c * width)
        self.rwrite(c * padside + " " * (content_width + 4) + c * padside)
        self.rwrite(c * padside + "  " + content_line + "  " + c * padside)
        self.rwrite(c * padside + " " * (content_width + 4) + c * padside)
        self.rwrite(c * width)
        self.rwrite(c * width)
        self.rwrite("")
        prefix = "Description: "
        preferredWidth = 80
        wrapper = textwrap.TextWrapper(
            initial_indent=prefix,
            width=preferredWidth,
            subsequent_indent=" " * len(prefix),
        )
        self.rwrite(wrapper.fill(self.vc_description))
        self.rwrite("")
        self.rwrite("Author: {}".format(self.vc_author))

    ############################################################################
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Leaving with block:
        Finishing VC calculation.  Send message to stdout, and close report file.
        """
        # Send message to stdout that VC calculation is starting
        print()
        print(
            "=== Verification check {} end ({}) ==="
            "".format(self.vc_name, time.strftime("%Y-%m-%d %H:%M:%S"))
        )
        # Close report file
        self.report.close()

    ############################################################################
    def write_results(self, results):
        """
        Use jinja2 to Process the template file 'results.edn.tpl' (expected
        in the current working directory) to replace the keys in 'results'.
        Place the new 'results.edn' file in a directory called 'output',
        creating it if it does not already exist.
        """
        if not self.report_ready:
            raise KIMVCError(
                "ERROR: Trying to write out results.edn "
                "but directory 'output' is not ready.\n"
            )

        # Turn lists into strings with double quotes to to be compatible with
        # EDN format
        for key in results:
            if isinstance(results[key], list):
                results[key] = str(results[key]).replace("'", '"')

        template_environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader("/"),
            block_start_string="@[",
            block_end_string="]@",
            variable_start_string="@<",
            variable_end_string=">@",
            comment_start_string="@#",
            comment_end_string="#@",
            undefined=jinja2.StrictUndefined,
        )
        # Template the EDN output
        with open(
            os.path.abspath(os.path.join("output", "results.edn")),
            "w",
            encoding="utf-8",
        ) as f:
            template = template_environment.get_template(
                os.path.abspath("results.edn.tpl")
            )
            f.write(template.render(**results))

    ############################################################################
    def write_aux_ase_atoms(self, aux_file, atoms, format):
        """
        Write the configuration in the ASE 'atoms' object to an auxiliary
        file 'aux_file' in the output directory in the specified 'format'.
        For supported formats, see https://wiki.fysik.dtu.dk/ase/ase/io/io.html
        """
        if not self.report_ready:
            raise KIMVCError(
                "ERROR: Trying to write out ASE Atoms "
                "aux file but directory 'output' is not ready.\n"
            )

        Atoms.write(atoms, os.path.join("output", aux_file), format=format)

    ############################################################################
    def write_aux_x_y(self, aux_file, x, y):
        """
        Write a file containing the lists x and y:
        x[0] y[0]
        x[1] y[1]
        ...
        If the lists are not the same length, the shorter sets the size.
        """
        if not self.report_ready:
            raise KIMVCError(
                "ERROR: Trying to write out x-y data "
                "aux file but directory 'output' is not ready.\n"
            )

        with open(
            os.path.abspath(os.path.join("output", aux_file)), "w", encoding="utf-8"
        ) as f:
            for i in range(0, min(len(x), len(y))):
                f.write("{0: 11.8e} {1: 11.8e}\n".format(x[i], y[i]))

    ############################################################################
    def write_aux_string(self, aux_file, string):
        """
        Write a file containing the contents in 'string'
        """
        if not self.report_ready:
            raise KIMVCError(
                "ERROR: Trying to write out string to "
                "aux file but directory 'output' is not ready.\n"
            )

        with open(
            os.path.abspath(os.path.join("output", aux_file)), "w", encoding="utf-8"
        ) as f:
            f.write("{}\n".format(string))


################################################################################
# If called directly, do nothing
if __name__ == "__main__":
    pass



"""
Simple wrapper for executable for converting arbitrary units to SI units

Copyright (c) 2014-2022, Regents of the University of Minnesota. All rights
reserved.

This software may be distributed as-is, without modification.
"""

import math
import re
import subprocess
import warnings

warnings.simplefilter("ignore")


class UnitConversion(Exception):
    """Class for unit conversion errors"""


_units_output_expression = re.compile(
    r"(?P<value>(?:[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?))(?: (?P<unit>.+))?"
)


def check_units_util():
    """
    Check that units util can be found
    """
    try:
        subprocess.check_output(["units", "--help"])
    except Exception:
        raise UnitConversion(
            "Failed to run a 'units' test command. It is likely "
            "that the 'units' executable was not found."
        )


def linear_fit(x, y):
    """
    Perform a linear fit between x,y, returning the average error for each data
    point as well. This is written this way so as to not add a numpy dependency
    """
    n = len(x)
    xx = sum([x**2 for x in x]) - sum(x) ** 2 / n
    xy = sum(map(lambda x, y: x * y, x, y)) - sum(x) * sum(y) / n
    a, b = sum(y) / n - xy / xx * sum(x) / n, xy / xx
    yhat = [a + b * x for x in x]
    yerr = math.sqrt(sum(map(lambda y, yh: (y - yh) ** 2 / y**2, y, yhat)) / n)
    return a, b, yerr


def islinear(unit, to_unit=None):
    """
    Detect if the conversion from `unit` to `to_unit` is a linear map. Apparently
    the units utility is float precision, so if error is less than 1e-7 we know
    it is linear.
    """
    x = [100 ** (1e-2 * (i - 50)) for i in range(20)]
    y = convert_list(x, unit, to_unit=to_unit, dofit=False)[0]
    a, b, err = linear_fit(x, y)

    a = convert_list(0, unit, to_unit=to_unit, dofit=False)[0]
    b = convert_list(1, unit, to_unit=to_unit, dofit=False)[0] - a
    return a, b, err < 1e-7


def convert_units(from_value, from_unit, wanted_unit=None, suppress_unit=False):
    """Works with 'units' utility"""
    check_units_util()
    from_sign = from_value < 0
    from_value = str(abs(from_value))
    from_unit = str(from_unit)

    TEMPERATURE_FUNCTION_UNITS = ["degC", "tempC", "degF", "tempF"]

    if from_unit in TEMPERATURE_FUNCTION_UNITS:
        args = [
            "units",
            "-o",
            "%1.15e",
            "-qt1",
            "".join((from_unit, "(", from_value, ")")),
        ]

    else:
        args = ["units", "-o", "%1.15e", "-qt1", " ".join((from_value, from_unit))]

    if wanted_unit:
        args.append(wanted_unit)

    try:
        output = subprocess.check_output(args).decode("utf-8")
    except subprocess.CalledProcessError:
        tag = wanted_unit if wanted_unit else "SI"
        raise UnitConversion(
            "Error in unit conversion of {} {} to {}".format(from_value, from_unit, tag)
        )

    matches = _units_output_expression.match(output).groupdict(None)
    out = ((-1) ** from_sign * float(matches["value"]), matches["unit"] or wanted_unit)

    if suppress_unit:
        return out[0]
    return out


# Set default behavior
convert = convert_units


def convert_list(x, from_unit, to_unit=None, convert=convert, dofit=True):
    """Thread conversion over a list, or list of lists"""
    # Need a list for scoping reasons

    # Constant shortcut
    if from_unit in (1, 1.0, "1"):
        to_unit = "1"

    # get the SI unit if none provided
    if to_unit is None:
        _, to_unit = convert(1.0, from_unit)

    def convert_inner(x, fit=None):
        if isinstance(x, (list, tuple)):
            return type(x)(convert_inner(i, fit=fit) for i in x)
        else:
            if to_unit == "1":
                return float(x)
            else:
                if fit is not None:
                    return fit[0] + fit[1] * x
                return float(convert(x, from_unit, to_unit, suppress_unit=True))

    # setup the linear fit if we are requested to simplify
    fit = None
    if dofit and isinstance(x, (list, tuple)) and len(x) > 20:
        a, b, linear = islinear(from_unit, to_unit)
        fit = (a, b) if linear else None

    output = convert_inner(x, fit=fit)
    return output, to_unit


def add_si_units(doc, convert=convert):
    """Given a document, add all of the appropriate si-units fields"""
    if isinstance(doc, dict):
        # check for a source-unit to defined a value with units
        if "source-unit" in doc:
            # we've found a place to add
            assert "source-value" in doc, "Badly formed doc"
            o_value = doc.get("source-value", None)
            o_unit = doc.get("source-unit", None)

            if o_value is None:
                raise UnitConversion("No source-value provided")
            if o_unit is None:
                raise UnitConversion("No source-unit provided")

            # convert the units and insert
            value, unit = convert_list(o_value, o_unit, convert=convert)
            si_dict = {"si-unit": unit, "si-value": value}
            doc = doc.copy()
            doc.update(si_dict)
            return doc
        else:
            # recurse
            return type(doc)(
                (key, add_si_units(value)) for key, value in list(doc.items())
            )

    elif isinstance(doc, (list, tuple)):
        return type(doc)(add_si_units(x) for x in doc)

    return doc



#!/usr/bin/python

import math

import numpy as np
from ase.build import bulk
from ase.calculators.kim.kim import KIM
from ase.io import read
from elastic_excerpt import get_unique_components_and_reconstruct_matrix
from sympy import matrix2numpy, symbols

from kim_tools import (
    CENTERING_DIVISORS,
    change_of_basis_atoms,
    get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice,
    get_crystal_structure_from_atoms,
    get_formal_bravais_lattice_from_space_group,
    get_space_group_number_from_prototype,
    get_symbolic_cell_from_formal_bravais_lattice,
)
from kim_tools.symmetry_util.core import (
    PeriodExtensionException,
    fit_voigt_tensor_to_cell_and_space_group,
    kstest_reduced_distances,
    reduce_and_avg,
)


def test_change_of_basis_atoms(
    atoms_conventional=bulk("SiC", "zincblende", 4.3596, cubic=True)
):
    calc = KIM("LJ_ElliottAkerson_2015_Universal__MO_959249795837_003")
    atoms_conventional.calc = calc
    crystal_structure = get_crystal_structure_from_atoms(atoms_conventional)
    prototype_label = crystal_structure["prototype-label"]["source-value"]
    sgnum = get_space_group_number_from_prototype(prototype_label)
    formal_bravais_lattice = get_formal_bravais_lattice_from_space_group(sgnum)
    primitive_to_conventional_change_of_basis = (
        get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice(
            formal_bravais_lattice
        )
    )
    conventional_to_primitive_change_of_basis = np.linalg.inv(
        primitive_to_conventional_change_of_basis
    )
    centering = formal_bravais_lattice[1]
    multiplier = np.linalg.det(primitive_to_conventional_change_of_basis)
    assert np.isclose(multiplier, CENTERING_DIVISORS[centering])
    conventional_energy = atoms_conventional.get_potential_energy()
    atoms_primitive = change_of_basis_atoms(
        atoms_conventional, conventional_to_primitive_change_of_basis
    )
    atoms_primitive.calc = calc
    primitive_energy = atoms_primitive.get_potential_energy()
    assert np.isclose(primitive_energy * multiplier, conventional_energy)
    atoms_conventional_rebuilt = change_of_basis_atoms(
        atoms_primitive, primitive_to_conventional_change_of_basis
    )
    atoms_conventional_rebuilt.calc = calc
    conventional_rebuilt_energy = atoms_conventional_rebuilt.get_potential_energy()
    assert np.isclose(conventional_energy, conventional_rebuilt_energy)


def test_test_reduced_distances():
    data_file_has_period_extension = {
        "structures/FeP_period_extended_phase_transition.data": True,
        "structures/FeP_stable.data": False,
    }
    repeat = [11, 11, 11]
    for data_file in data_file_has_period_extension:
        has_period_extension = data_file_has_period_extension[data_file]
        atoms = read(data_file, format="lammps-data")
        _, reduced_distances = reduce_and_avg(atoms, repeat)
        try:
            kstest_reduced_distances(reduced_distances)
            assert not has_period_extension
        except PeriodExtensionException:
            assert has_period_extension


def test_fit_voigt_tensor_to_cell_and_space_group():
    a, b, c, alpha, beta, gamma = symbols("a b c alpha beta gamma")
    # taken from A5B11CD8E_aP26_1_5a_11a_a_8a_a-001
    test_substitutions = [
        (a, 1.0),
        (b, 1.20466246551),
        (c, 1.81123604761),
        (alpha, math.radians(76.515)),
        (beta, math.radians(81.528)),
        (gamma, math.radians(71.392)),
    ]
    # Generate a random symmetric matrix
    c = np.random.random((6, 6))
    c = c + c.T
    for sgnum in range(1, 231):
        lattice = get_formal_bravais_lattice_from_space_group(sgnum)
        symbolic_cell = get_symbolic_cell_from_formal_bravais_lattice(lattice)
        cell = matrix2numpy(symbolic_cell.subs(test_substitutions), dtype=float)
        # This takes any matrix, picks out the unique constants based on the
        # algebraic diagrams, and returns a matrix conforming to the material symmetry

        # Test 1: fit_voigt_tensor_to_cell_and_space_group should not change
        # a matrix that is already symmetrized
        _, _, c_sym_alg = get_unique_components_and_reconstruct_matrix(c, sgnum)
        c_sym_alg_tens = fit_voigt_tensor_to_cell_and_space_group(
            c_sym_alg, cell, sgnum
        )
        assert np.allclose(c_sym_alg, c_sym_alg_tens)

        # Test 2: fit_voigt_tensor_to_cell_and_space_group should produce
        # a matrix that is already symmetrized
        c_sym_tens = fit_voigt_tensor_to_cell_and_space_group(c, cell, sgnum)
        if lattice == "aP":
            assert np.allclose(c, c_sym_tens)
        _, _, c_sym_tens_alg = get_unique_components_and_reconstruct_matrix(
            c_sym_tens, sgnum
        )
        assert np.allclose(c_sym_tens, c_sym_tens_alg)


if __name__ == "__main__":
    test_fit_voigt_tensor_to_cell_and_space_group()



#!/usr/bin/python

import json
import logging
import os
from os import PathLike
from random import random
from typing import Dict, List, Optional

import numpy as np
import numpy.typing as npt
import pytest
from ase import Atoms
from ase.calculators.kim.kim import KIM

from kim_tools import (
    AFLOW,
    detect_unique_crystal_structures,
    get_atoms_from_crystal_structure,
    get_bravais_lattice_from_prototype,
    get_real_to_virtual_species_map,
    get_space_group_number_from_prototype,
    get_wyckoff_lists_from_prototype,
    minimize_wrapper,
    prototype_labels_are_equivalent,
    query_crystal_structures,
    split_parameter_array,
)
from kim_tools.aflow_util.core import (
    build_abstract_formula_from_stoich_reduced_list,
    find_species_permutation_between_prototype_labels,
)

logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)

TEST_CASES = [572, 365, 1729, 1194, 1473, 166, 1205, 1357, 915, 212, 641, 22]
MATERIALS_FILE = "test_structures.json"
QUERY_DUMP = "output/query_result.json"


def shuffle_atoms(atoms: Atoms) -> Atoms:
    atoms_shuffled = atoms.copy()
    permute = np.random.permutation(len(atoms_shuffled))
    atoms_shuffled.set_scaled_positions(
        [atoms.get_scaled_positions()[i] for i in permute]
    )
    atoms_shuffled.set_chemical_symbols(
        [atoms.get_chemical_symbols()[i] for i in permute]
    )
    return atoms_shuffled


def _frac_pos_match_sanity_checks(
    reference_positions: npt.ArrayLike,
    test_positions: npt.ArrayLike,
    reference_species: Optional[List] = None,
    test_species: Optional[List] = None,
) -> bool:
    """
    Sanity checks for comparing sets of fractional positions
    """
    if reference_species is None or test_species is None:
        if not (reference_species is None and test_species is None):
            logger.warning(
                "Refusing to compare positions when one structure has species given "
                "and the other does not"
            )
            return False
        logger.info("Comparing fractional positions without species")
    else:
        if not (
            len(reference_positions)
            == len(test_positions)
            == len(reference_species)
            == len(test_species)
        ):
            logger.info(
                "Atomic positions and/or species lists have different lengths between "
                "test and reference"
            )
            return False

    if len(reference_positions) != len(test_positions):
        logger.info("Number of atomic positions does not match")
        return False
    return True


def frac_pos_match_allow_permute_wrap(
    reference_positions: npt.ArrayLike,
    test_positions: npt.ArrayLike,
    reference_species: Optional[List] = None,
    test_species: Optional[List] = None,
) -> bool:
    """
    Check if fractional positions match allowing for permutations and PBC wrapping
    """
    if not _frac_pos_match_sanity_checks(
        reference_positions, test_positions, reference_species, test_species
    ):
        return False

    test_position_matched = [False] * len(test_positions)
    for i, reference_position in enumerate(reference_positions):
        for j, test_position in enumerate(test_positions):
            if test_position_matched[j]:  # this position already matched.
                continue
            position_differences = np.asarray(reference_position) - np.asarray(
                test_position
            )
            if np.allclose(
                position_differences, np.rint(position_differences), atol=1e-5
            ):
                if reference_species is not None:
                    if reference_species[i] != test_species[j]:
                        logger.info(
                            f"Reference position {i} matches test position {j} but the "
                            "species do not."
                        )
                        return False
                test_position_matched[j] = True
                break

    if all(test_position_matched):
        logger.info("Successfully matched the fractional positions")
        return True
    else:
        logger.info("Not all fractional positions successfully matched")
        return False


def frac_pos_match_allow_wrap(
    reference_positions: npt.ArrayLike,
    test_positions: npt.ArrayLike,
    reference_species: Optional[List] = None,
    test_species: Optional[List] = None,
) -> bool:
    """
    Check if fractional positions match allowing for PBC wrapping but maintaining order
    """
    if not _frac_pos_match_sanity_checks(
        reference_positions, test_positions, reference_species, test_species
    ):
        return False

    if reference_species is not None:
        if reference_species != test_species:
            logger.info(
                f"Species lists do not match. Got\n{test_species}\nexpected\n"
                + str(reference_species)
            )
            return False

    for ref_pos, test_pos in zip(reference_positions, test_positions):
        position_differences = np.asarray(ref_pos) - np.asarray(test_pos)
        if not np.allclose(
            position_differences, np.rint(position_differences), atol=1e-5
        ):
            logger.info(f"Failed to match positions, expected {ref_pos} got {test_pos}")
            return False

    logger.info("Successfully matched the fractional positions")
    return True


def atoms_frac_pos_match_allow_permute_wrap(reference_atoms: Atoms, test_atoms: Atoms):
    return frac_pos_match_allow_permute_wrap(
        reference_atoms.get_scaled_positions(),
        test_atoms.get_scaled_positions(),
        reference_atoms.get_chemical_symbols(),
        test_atoms.get_chemical_symbols(),
    )


def atoms_frac_pos_match_allow_wrap(reference_atoms: Atoms, test_atoms: Atoms):
    return frac_pos_match_allow_wrap(
        reference_atoms.get_scaled_positions(),
        test_atoms.get_scaled_positions(),
        reference_atoms.get_chemical_symbols(),
        test_atoms.get_chemical_symbols(),
    )


def get_test_crystal_structures(
    materials_file: PathLike = MATERIALS_FILE,
    test_cases: Optional[List[int]] = TEST_CASES,
    deduplicate: bool = False,
) -> List[Dict]:
    """
    Query OpenKIM reference data for materials (prototype label + species) from
    `materials_file`. Deduplicate the structures if asked, otherwise just return the
    first index of each (you don't want to be testing 25 identical copies of FCC
    Aluminum)

    Returns a list of `crystal-structure-npt` property instances
    """
    with open(materials_file) as f:
        query_inputs = json.load(f)

    if test_cases is not None:
        query_inputs = [query_inputs[i] for i in test_cases]

    test_crystal_structures = []
    for query_input in query_inputs:
        query_result = query_crystal_structures(**query_input)
        if deduplicate:
            indices_to_test = detect_unique_crystal_structures(query_result)
        else:
            indices_to_test = [0]
        test_crystal_structures += [query_result[i] for i in indices_to_test]

    with open(QUERY_DUMP, "w") as f:
        json.dump(test_crystal_structures, f)

    return test_crystal_structures


@pytest.fixture(scope="module")
def input_crystal_structures():
    return get_test_crystal_structures()


def test_prototype_labels_are_equivalent():
    assert not (prototype_labels_are_equivalent("AB_oC8_65_j_g", "AB_oC8_65_i_i"))
    assert prototype_labels_are_equivalent("AB2C_oP8_17_a_bc_d", "AB2C_oP8_17_a_bd_c")
    assert prototype_labels_are_equivalent("AB_mP8_14_ad_e", "AB_mP8_14_ab_e")
    assert prototype_labels_are_equivalent("AB_mP8_14_ad_e", "AB_mP8_14_ab_e-001")
    assert find_species_permutation_between_prototype_labels(
        "AB2C_oP8_17_a_bc_d", "ABC2_oP8_17_a_c_bd"
    ) == (0, 2, 1)
    assert (
        find_species_permutation_between_prototype_labels(
            "AB2C_oP8_17_a_bc_d", "AB2C_oP8_17_a_c_bd"
        )
        is None
    )


def test_build_abstract_formula_from_stoich_reduced_list():
    assert build_abstract_formula_from_stoich_reduced_list([1, 2, 3]) == "AB2C3"


def test_get_wyckoff_lists_from_prototype():
    assert get_wyckoff_lists_from_prototype("A_hP68_194_ef2h2kl") == ["efhhkkl"]
    assert get_wyckoff_lists_from_prototype("AB_mC48_8_12a_12a") == [
        "aaaaaaaaaaaa",
        "aaaaaaaaaaaa",
    ]


def test_get_param_names_from_prototype():
    aflow = AFLOW()
    assert aflow.get_param_names_from_prototype("A_cF4_225_a") == ["a"]
    assert aflow.get_param_names_from_prototype("A_cF240_202_h2i") == [
        "a",
        "y1",
        "z1",
        "x2",
        "y2",
        "z2",
        "x3",
        "y3",
        "z3",
    ]


def test_get_equations_from_prototype(input_crystal_structures):
    aflow = AFLOW()
    # for large-scale testing, helpful to check that same prototype with different
    # parameters gives the same results
    equation_sets_cache = {}
    for material in input_crystal_structures:
        species = material["stoichiometric-species"]["source-value"]
        real_to_virtual_species_map = get_real_to_virtual_species_map(species)
        prototype_label = material["prototype-label"]["source-value"]
        parameter_names = aflow.get_param_names_from_prototype(prototype_label)
        _, internal_parameter_names_ref = split_parameter_array(parameter_names)
        parameter_values = [material["a"]["source-value"]]
        if "parameter-values" in material:
            parameter_values += material["parameter-values"]["source-value"]
        atoms = get_atoms_from_crystal_structure(material)
        if prototype_label not in equation_sets_cache:
            equation_sets = aflow.get_equation_sets_from_prototype(prototype_label)
            equation_sets_cache[prototype_label] = equation_sets
        else:
            equation_sets = equation_sets_cache[prototype_label]

        internal_parameter_names = []
        for eqset in equation_sets:
            internal_parameter_names += eqset.param_names

        # TODO: Does it matter that the order is not the same?
        assert set(internal_parameter_names) == set(internal_parameter_names_ref), (
            "get_equation_sets_from_prototype got incorrect internal parameter names, "
            f"got {internal_parameter_names} expected {internal_parameter_names_ref}"
        )

        assert sum([len(eqset.coeff_matrix_list) for eqset in equation_sets]) == len(
            atoms
        ), "get_equation_sets_from_prototype got an incorrect number of equations"

        scaled_positions_computed_from_equations = []
        virtual_species_from_atoms = []
        species_from_equations = []

        diagnostics = (
            f"Problem occurred in prototype {prototype_label}\n"
            "Reference fractional positions and species:\n"
        )
        for atom in atoms:
            for position in atom.scaled_position:
                diagnostics += f"{position:8.4f}"
            virtual_species = real_to_virtual_species_map[atom.symbol]
            diagnostics += f"    {virtual_species}\n"
            virtual_species_from_atoms.append(virtual_species)
        diagnostics += "\nComputed fractional positions and species:\n"

        for eqset in equation_sets:
            for coeff_mat, const_terms in zip(
                eqset.coeff_matrix_list, eqset.const_terms_list
            ):
                scaled_position = (
                    coeff_mat
                    @ [
                        [parameter_values[parameter_names.index(parname)]]
                        for parname in eqset.param_names
                    ]
                    + const_terms
                )
                scaled_positions_computed_from_equations.append(scaled_position.T)
                species_from_equations.append(eqset.species)
                for position in scaled_position:
                    diagnostics += f"{position[0]:8.4f}"
                diagnostics += f"    {eqset.species}\n"

        assert frac_pos_match_allow_permute_wrap(
            atoms.get_scaled_positions(),
            scaled_positions_computed_from_equations,
            virtual_species_from_atoms,
            species_from_equations,
        ), f"Failed to match fractional coordinates.\n{diagnostics}"

        assert frac_pos_match_allow_wrap(
            atoms.get_scaled_positions(),
            scaled_positions_computed_from_equations,
            virtual_species_from_atoms,
            species_from_equations,
        ), (
            "Matched fractional coordinates, but there was a permutation.\n"
            + diagnostics
        )

        print(
            f"Successfully checked get_equations_from_prototype for {prototype_label}"
        )


def test_solve_for_params_of_known_prototype(input_crystal_structures):
    aflow = AFLOW(np=19)
    match_counts_by_bravais_lattice = {}
    match_counts_by_spacegroup = {}
    INIT_COUNTS = {"match": 0, "nonmatch": 0}

    for bravais_lattice in [
        "aP",
        "mP",
        "mC",
        "oP",
        "oC",
        "oF",
        "oI",
        "tP",
        "tI",
        "hP",
        "hR",
        "cP",
        "cF",
        "cI",
    ]:
        match_counts_by_bravais_lattice[bravais_lattice] = INIT_COUNTS.copy()
    for spacegroup in range(1, 231):
        match_counts_by_spacegroup[spacegroup] = INIT_COUNTS.copy()

    failed_to_solve_at_least_one = False

    for material in input_crystal_structures:
        prototype_label = material["prototype-label"]["source-value"]

        bravais_lattice = get_bravais_lattice_from_prototype(prototype_label)

        bravais_lattice = get_bravais_lattice_from_prototype(prototype_label)
        spacegroup = get_space_group_number_from_prototype(prototype_label)

        atoms = get_atoms_from_crystal_structure(material)

        atoms = shuffle_atoms(atoms)

        atoms.rotate(
            (random(), random(), random()),
            (random(), random(), random()),
            rotate_cell=True,
        )

        atoms.translate((random() * 1000, random() * 1000, random() * 1000))

        atoms.wrap()

        atoms.calc = KIM("LJ_ElliottAkerson_2015_Universal__MO_959249795837_003")
        minimize_wrapper(
            atoms,
            fix_symmetry=True,
            variable_cell=True,
            steps=2,
            opt_kwargs={"maxstep": 0.05},
        )

        try:
            aflow.solve_for_params_of_known_prototype(atoms, prototype_label)
            crystal_did_not_rotate = True
        except Exception as e:
            crystal_did_not_rotate = False
            print(e)

        if not crystal_did_not_rotate:
            failed_to_solve_at_least_one = True
            print(
                "Failed to solve for parameters or confirm unrotated prototype "
                f"designation for {prototype_label}"
            )
            match_counts_by_bravais_lattice[bravais_lattice]["nonmatch"] += 1
            match_counts_by_spacegroup[spacegroup]["nonmatch"] += 1
            filename = f"output/{prototype_label}.POSCAR"
            filename_exists = os.path.isfile(filename)
            if filename_exists:
                suffix = 0
                while not filename_exists:
                    filename = f"output/{prototype_label}.{suffix}.POSCAR"
                    filename_exists = os.path.isfile(filename)
                    suffix += 1
            print(f"Dumping atoms to {filename}")
            atoms.write(filename, format="vasp", sort=True)
        else:
            print(
                "Successfully confirmed unrotated prototype designation for "
                + prototype_label
            )
            match_counts_by_bravais_lattice[bravais_lattice]["match"] += 1
            match_counts_by_spacegroup[spacegroup]["match"] += 1
    with open("output/match_counts_by_bravais_lattice.json", "w") as f:
        json.dump(match_counts_by_bravais_lattice, f)
    with open("output/match_counts_by_spacegroup.json", "w") as f:
        json.dump(match_counts_by_spacegroup, f)

    assert not failed_to_solve_at_least_one


if __name__ == "__main__":
    with open("output.3/query_result.json") as f:
        test_solve_for_params_of_known_prototype(json.load(f))



"""Tools for working with crystal prototypes using the AFLOW command line tool"""

import json
import logging
import os
import subprocess
from curses.ascii import isalpha, isdigit
from dataclasses import dataclass
from itertools import permutations
from math import acos, cos, degrees, radians, sqrt
from os import PathLike
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import ase
import numpy as np
import numpy.typing as npt
from ase import Atoms
from ase.cell import Cell
from ase.neighborlist import natural_cutoffs, neighbor_list
from semver import Version
from sympy import Symbol, linear_eq_to_matrix, matrix2numpy, parse_expr

from ..symmetry_util import (
    A_CENTERED_ORTHORHOMBIC_GROUPS,
    C_CENTERED_ORTHORHOMBIC_GROUPS,
    CENTERING_DIVISORS,
    IncorrectNumAtomsException,
    are_in_same_wyckoff_set,
    cartesian_rotation_is_in_point_group,
    get_possible_primitive_shifts,
    get_primitive_wyckoff_multiplicity,
    get_wyck_pos_xform_under_normalizer,
    space_group_numbers_are_enantiomorphic,
)

logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)


__author__ = ["ilia Nikiforov", "Ellad Tadmor"]
__all__ = [
    "EquivalentEqnSet",
    "EquivalentAtomSet",
    "write_tmp_poscar_from_atoms_and_run_function",
    "get_equivalent_atom_sets_from_prototype_and_atom_map",
    "IncorrectSpaceGroupException",
    "IncorrectSpeciesException",
    "InconsistentWyckoffException",
    "check_number_of_atoms",
    "split_parameter_array",
    "internal_parameter_sort_key",
    "get_stoich_reduced_list_from_prototype",
    "get_wyckoff_lists_from_prototype",
    "prototype_labels_are_equivalent",
    "get_space_group_number_from_prototype",
    "get_pearson_symbol_from_prototype",
    "get_bravais_lattice_from_prototype",
    "read_shortnames",
    "get_real_to_virtual_species_map",
    "solve_for_aflow_cell_params_from_primitive_ase_cell_params",
    "AFLOW",
]

AFLOW_EXECUTABLE = "aflow"
REQUIRED_AFLOW = "4.0.5"
AFLOW_PROTOTYPE_ENCYCLOPEDIA_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "aflow_prototype_encyclopedia"
)


class IncorrectSpaceGroupException(Exception):
    """
    Raised when spglib or aflow --sgdata detects a different space group than the one
    specified in the prototype label
    """


class IncorrectSpeciesException(Exception):
    """
    Raised when number or identity of species is inconsistent
    """


class InconsistentWyckoffException(Exception):
    """
    Raised when an insonsistency in Wyckoff positions is detected
    """


@dataclass
class EquivalentEqnSet:
    """
    Set of equations representing the fractional positions of equivalent atoms
    """

    species: str
    wyckoff_letter: str
    # The n free parameters associated with this Wyckoff postition, 0 <= n <= 3
    param_names: List[str]
    # m x 3 x n matrices of coefficients, where m is the multiplicity of the Wyckoff
    # position
    coeff_matrix_list: List[npt.ArrayLike]
    # m x 3 x 1 columns of constant terms in the coordinates. This gets subtracted from
    # the RHS when solving
    const_terms_list: List[npt.ArrayLike]


@dataclass
class EquivalentAtomSet:
    """
    Set of equivalent atoms
    """

    species: str
    wyckoff_letter: str
    frac_position_list: List[npt.ArrayLike]  # m x 3 x 1 columns


def write_tmp_poscar_from_atoms_and_run_function(
    atoms: Atoms, function: Callable, *args, **kwargs
) -> Any:
    """
    Write the Atoms file to a NamedTemporaryFile and run 'function' on it.

    Args:
        atoms:
            The atoms object that will be written to a POSCAR file and fed as the
            first argument to function
        function: A function that takes a POSCAR file as the first argument

    Returns:
        Whatever `function` returns
    """
    with NamedTemporaryFile("w+") as fp:
        atoms.write(fp, sort=True, format="vasp")
        fp.seek(0)
        return function(fp.name, *args, **kwargs)


def check_number_of_atoms(
    atoms: Atoms, prototype_label: str, primitive_cell: bool = True
) -> None:
    """
    Check if the Atoms object (which must be a conventional or primitive unit cell)
    has the correct number of atoms according to prototype_label

    Raises:
        IncorrectNumAtomsException
    """
    prototype_label_list = prototype_label.split("_")
    pearson = prototype_label_list[1]

    # get the number of atoms in conventional cell from the Pearson symbol
    num_conv_cell = 0
    for character in pearson:
        if character.isdigit():
            num_conv_cell *= 10
            num_conv_cell += int(character)

    centering = pearson[1]

    if centering == "R":
        num_conv_cell *= 3

    if not primitive_cell:
        num_lattice = 1
    else:
        num_lattice = CENTERING_DIVISORS[centering]

    # This check is probably really extraneous, but better safe than sorry
    if num_conv_cell % num_lattice != 0:
        raise IncorrectNumAtomsException(
            f"WARNING: Number of atoms in conventional cell {num_conv_cell} derived "
            f"from Pearson symbol of prototype {prototype_label} is not divisible by "
            f"the number of lattice points {num_lattice}"
        )

    num_cell = num_conv_cell / num_lattice

    if len(atoms) != num_cell:
        raise IncorrectNumAtomsException(
            f"WARNING: Number of ASE atoms {len(atoms)} does not match Pearson symbol "
            f" of prototype {prototype_label}"
        )


def split_parameter_array(
    parameter_names: List[str], list_to_split: Optional[List] = None
) -> Tuple[List, List]:
    """
    Split a list of parameters into cell and internal parameters.

    Args:
        parameter_names:
            List of AFLOW parameter names, e.g.
            `["a", "c/a", "x1", "x2", "y2", "z2"]`
            Proper AFLOW order is assumed, i.e. cell parameters
            first, then internal.
        list_to_split:
            List to split, must be same length as `parameter_names`
            If omitted, `parameter_names` itself will be split

    Returns:
        `list_to_split` (or `parameter_names` if `list_to_split` is omitted),
        split into lists corresponding to the split between cell and internal parameters

    Raises:
        AssertionError:
            If lengths are incompatible or if `parameter_names` fails an (incomplete)
            check that it is a sensible list of AFLOW parameters
    """
    if list_to_split is None:
        list_to_split = parameter_names

    assert len(list_to_split) == len(
        parameter_names
    ), "`list_to_split` must have the same length as `parameter_names`"

    in_internal_part = False

    cell_part = []
    internal_part = []

    CARTESIAN_AXES = ["x", "y", "z"]

    for name, value in zip(parameter_names, list_to_split):
        assert isinstance(
            name, str
        ), "At least one element of `parameter_names` is not a string."
        if not in_internal_part:
            if name[0] in CARTESIAN_AXES:
                in_internal_part = True
        else:
            # means we have already encountered
            # an internal coordinate in a past iteration
            assert name[0] in CARTESIAN_AXES, (
                "`parameter_names` seems to have an internal parameter "
                "followed by a non-internal one"
            )

        if in_internal_part:
            internal_part.append(value)
        else:
            cell_part.append(value)

    return cell_part, internal_part


def internal_parameter_sort_key(parameter_name: Union[Symbol, str]) -> int:
    """
    Sorting key for internal free parameters. Sort by number first, then letter
    """
    parameter_name_str = str(parameter_name)
    axis = parameter_name_str[0]
    assert (
        axis == "x" or axis == "y" or axis == "z"
    ), "Parameter name must start with x, y, or z"
    number = int(parameter_name_str[1:])
    return 1000 * number + ord(axis)


def get_equivalent_atom_sets_from_prototype_and_atom_map(
    atoms: Atoms, prototype_label: str, atom_map: List[int], sort_atoms: bool = False
) -> List[EquivalentAtomSet]:
    """
    Get a list of objects representing sets of equivalent atoms from the atom_map of an
    AFLOW comparison

    The AFLOW comparison should be between `atoms` and `atoms_rebuilt`, in that order,
    where `atoms_rebuilt` is regenerated from the prototype designation detected from
    `atoms`, and `prototype_label` is the detected prototype label

    Args:
        sort_atoms:
            If `atom_map` was obtained by sorting `atoms` before writing it to
            POSCAR, set this to True
    """

    if sort_atoms:
        with NamedTemporaryFile("w+") as f:
            atoms.write(f.name, format="vasp", sort=True)
            f.seek(0)
            atoms_in_correct_order = ase.io.read(f.name, format="vasp")
    else:
        atoms_in_correct_order = atoms

    # initialize return list
    equivalent_atom_set_list = []

    redetected_wyckoff_lists = get_wyckoff_lists_from_prototype(prototype_label)
    sg_num = get_space_group_number_from_prototype(prototype_label)

    index_in_atoms_rebuilt = 0
    for i_species, species_wyckoff_list in enumerate(redetected_wyckoff_lists):
        virtual_species_letter = chr(65 + i_species)
        for wyckoff_letter in species_wyckoff_list:
            equivalent_atom_set_list.append(
                EquivalentAtomSet(virtual_species_letter, wyckoff_letter, [])
            )
            for _ in range(get_primitive_wyckoff_multiplicity(sg_num, wyckoff_letter)):
                # atom_map[index_in_atoms] = index_in_atoms_rebuilt
                index_in_atoms = atom_map.index(index_in_atoms_rebuilt)
                equivalent_atom_set_list[-1].frac_position_list.append(
                    atoms_in_correct_order.get_scaled_positions()[
                        index_in_atoms
                    ].reshape(3, 1)
                )
                index_in_atoms_rebuilt += 1

    return equivalent_atom_set_list


def get_stoich_reduced_list_from_prototype(prototype_label: str) -> List[int]:
    """
    Get numerical list of stoichiometry from prototype label, i.e. "AB3_hP8..." -> [1,3]

    Args:
        prototype_label:
            AFLOW prototype label

    Returns:
        List of reduced stoichiometric numbers
    """
    stoich_reduced_formula = prototype_label.split("_")[0]
    stoich_reduced_list = []
    stoich_reduced_curr = None
    for char in stoich_reduced_formula:
        if isalpha(char):
            if stoich_reduced_curr is not None:
                if stoich_reduced_curr == 0:
                    stoich_reduced_curr = 1
                stoich_reduced_list.append(stoich_reduced_curr)
            stoich_reduced_curr = 0
        else:
            assert isdigit(char)
            # will throw an error if we haven't encountered an alphabetical letter, good
            stoich_reduced_curr *= 10
            stoich_reduced_curr += int(char)
    # write final number
    if stoich_reduced_curr == 0:
        stoich_reduced_curr = 1
    stoich_reduced_list.append(stoich_reduced_curr)
    return stoich_reduced_list


def build_abstract_formula_from_stoich_reduced_list(
    stoich_reduced_list: List[int],
) -> str:
    """
    Get abstract chemical formula from numerical list of stoichiometry
    i.e. [1,3] -> "AB3"

    Args:
        stoich_reduced_list:
            List of reduced stoichiometric numbers

    Returns:
        Abstract chemical formula
    """
    formula = ""
    for spec_num, stoich_num in enumerate(stoich_reduced_list):
        assert isinstance(stoich_num, int)
        assert stoich_num > 0
        formula += chr(65 + spec_num)
        if stoich_num > 1:
            formula += str(stoich_num)
    return formula


def get_wyckoff_lists_from_prototype(prototype_label: str) -> List[str]:
    """
    Expand the list of Wyckoff letters in the prototype to account for each individual
    letter instead of using numerical multipliers for repeated letters.
    e.g. A2B3C_mC48_15_aef_3f_2e -> ['aef','fff','ee']
    """
    expanded_wyckoff_letters = []
    prototype_label_split = prototype_label.split("-")[0].split("_")
    for species_wyckoff_string in prototype_label_split[3:]:
        expanded_wyckoff_letters.append("")
        curr_wyckoff_count = 0
        for char in species_wyckoff_string:
            if isalpha(char):
                if curr_wyckoff_count == 0:
                    curr_wyckoff_count = 1
                expanded_wyckoff_letters[-1] += char * curr_wyckoff_count
                curr_wyckoff_count = 0
            else:
                assert isdigit(char)
                curr_wyckoff_count *= 10  # if it's zero, we're all good
                curr_wyckoff_count += int(char)
    return expanded_wyckoff_letters


def prototype_labels_are_equivalent(
    prototype_label_1: str,
    prototype_label_2: str,
    allow_enantiomorph: bool = False,
) -> bool:
    """
    Checks if two prototype labels are equivalent (species permutations not allowed)

    Args:
        allow_enantiomorph:
            Whether to consider enantiomorphic pairs of space groups to be equivalent
    """

    if prototype_label_1 == prototype_label_2:
        return True

    # Check stoichiometry
    stoich_reduced_list_1 = get_stoich_reduced_list_from_prototype(prototype_label_1)

    if not stoich_reduced_list_1 == get_stoich_reduced_list_from_prototype(
        prototype_label_2
    ):
        logger.info(
            "Found non-matching stoichiometry in labels "
            f"{prototype_label_1} and {prototype_label_2}"
        )
        return False

    # Check Pearson symbol
    if not get_pearson_symbol_from_prototype(
        prototype_label_1
    ) == get_pearson_symbol_from_prototype(prototype_label_2):
        logger.info(
            "Found non-matching Pearson symbol in labels "
            f"{prototype_label_1} and {prototype_label_2}"
        )
        return False

    # Check space group number
    sg_num_1 = get_space_group_number_from_prototype(prototype_label_1)
    sg_num_2 = get_space_group_number_from_prototype(prototype_label_2)
    if allow_enantiomorph and not space_group_numbers_are_enantiomorphic(
        sg_num_2, sg_num_1
    ):
        logger.info(
            "Found non-matching Space group in labels "
            f"{prototype_label_1} and {prototype_label_2}"
        )
        return False
    elif sg_num_2 != sg_num_1:
        logger.info(
            "Found non-matching Space group in labels "
            f"{prototype_label_1} and {prototype_label_2}"
        )
        return False

    # OK, so far everything matches, now check the Wyckoff letters
    # Get lists of Wyckoff letters for each species,
    # e.g. A2B3C_mC48_15_aef_3f_2e -> ['aef', 'fff', 'ee']
    wyckoff_lists_1 = get_wyckoff_lists_from_prototype(prototype_label_1)
    wyckoff_lists_2 = get_wyckoff_lists_from_prototype(prototype_label_2)
    num_species = len(stoich_reduced_list_1)
    assert len(wyckoff_lists_1) == len(wyckoff_lists_2) == num_species, (
        "Somehow I got non-matching lists of Wyckoff letters, "
        "the prototype labels are probably malformed"
    )

    if sg_num_1 >= 16:
        # Theoretically, unless we are allowing species permutations, orthorhombic and
        # higher SGs should always have identical prototype labels due to minimal
        # Wyckoff enumeration. However, there are bugs in AFLOW making this untrue.

        wyck_pos_xform_list = get_wyck_pos_xform_under_normalizer(sg_num_1)
        for wyck_pos_xform in wyck_pos_xform_list:
            wyckoff_lists_match_for_each_species = True
            for i in range(num_species):
                wyckoff_list_ref = wyckoff_lists_1[i]
                wyckoff_list_test = ""
                for test_letter in wyckoff_lists_2[i]:
                    if test_letter == "A":
                        # SG47 runs out of the alphabet and uses capital A for its
                        # general position (which is unchanged under any transformation
                        # in the normalizer). However, we need "A" to be last, so we
                        # make it "{" to make it sort after all lowercase letters and
                        # replace it after
                        wyckoff_list_test += "{"
                    else:
                        test_letter_index = ord(test_letter) - 97
                        wyckoff_list_test += wyck_pos_xform[test_letter_index]
                wyckoff_list_test = "".join(sorted(wyckoff_list_test))
                wyckoff_list_test = wyckoff_list_test.replace("{", "A")
                if wyckoff_list_test != wyckoff_list_ref:
                    wyckoff_lists_match_for_each_species = False
                    break
            if wyckoff_lists_match_for_each_species:
                logger.warning(
                    f"Labels {prototype_label_1} and {prototype_label_2} were found to "
                    "be equivalent despite being non-identical. "
                    "This indicates a failure to find the lowest Wyckoff enumeration."
                )
                return True
        logger.info(
            f"Labels {prototype_label_1} and {prototype_label_2} were not found to be "
            "equivalent under any permutations allowable by the normalizer."
        )
        return False
    else:
        for wyckoff_list_1, wyckoff_list_2 in zip(wyckoff_lists_1, wyckoff_lists_2):
            for letter_1, letter_2 in zip(wyckoff_list_1, wyckoff_list_2):
                # Wyckoff sets are alphabetically contiguous in SG1-15, there is no
                # need to re-sort anything.
                # This is NOT true for all SGs (e.g. #200, Wyckoff set eh )
                if not are_in_same_wyckoff_set(letter_1, letter_2, sg_num_1):
                    logger.info(
                        f"Labels {prototype_label_1} and {prototype_label_2} have "
                        f"corresponding letters {letter_1} and {letter_2} that are not "
                        "in the same Wyckoff set"
                    )
                    return False
        logger.info(
            f"Labels {prototype_label_1} and {prototype_label_2} were found to be "
            "equivalent despite being non-identical. This is a normal occurrence for "
            "triclinic and monoclinic space groups such as this."
        )
        return True


def find_species_permutation_between_prototype_labels(
    prototype_label_1: str,
    prototype_label_2: str,
    allow_enantiomorph: bool = False,
) -> Optional[Tuple[int]]:
    """
    Find the permutation of species required to match two prototype labels

    Args:
        allow_enantiomorph:
            Whether to consider enantiomorphic pairs of space groups to be equivalent

    Returns:
        The permutation of species of ``prototype_label_1`` required to match
        ``prototype_label_2``, or None if no match is found
    """
    # Disassemble prototype_label_1
    stoich_reduced_list_1 = get_stoich_reduced_list_from_prototype(prototype_label_1)
    pearson_1 = get_pearson_symbol_from_prototype(prototype_label_1)
    space_group_1 = str(get_space_group_number_from_prototype(prototype_label_1))
    species_wyckoff_sections_1 = prototype_label_1.split("-")[0].split("_")[3:]

    nspecies = len(stoich_reduced_list_1)
    assert nspecies == len(species_wyckoff_sections_1)

    permutation_candidates = permutations(tuple(range(nspecies)))
    for permutation in permutation_candidates:
        # Permute the species
        stoich_reduced_list_1_permuted = [stoich_reduced_list_1[i] for i in permutation]
        species_wyckoff_sections_1_permuted = [
            species_wyckoff_sections_1[i] for i in permutation
        ]

        # Reassemble prototype_label_1_permuted
        abstract_formula_1_permuted = build_abstract_formula_from_stoich_reduced_list(
            stoich_reduced_list_1_permuted
        )
        prototype_label_1_permuted_list = [
            abstract_formula_1_permuted,
            pearson_1,
            space_group_1,
        ] + species_wyckoff_sections_1_permuted
        prototype_label_1_permuted = "_".join(prototype_label_1_permuted_list)
        if prototype_labels_are_equivalent(
            prototype_label_1=prototype_label_1_permuted,
            prototype_label_2=prototype_label_2,
            allow_enantiomorph=allow_enantiomorph,
        ):
            return permutation
    return None


def get_space_group_number_from_prototype(prototype_label: str) -> int:
    return int(prototype_label.split("_")[2])


def get_pearson_symbol_from_prototype(prototype_label: str) -> str:
    return prototype_label.split("_")[1]


def get_bravais_lattice_from_prototype(prototype_label: str) -> str:
    return get_pearson_symbol_from_prototype(prototype_label)[:2]


def read_shortnames(
    aflow_prototype_encyclopedia_path: PathLike = AFLOW_PROTOTYPE_ENCYCLOPEDIA_PATH,
) -> Dict:
    """
    Read the aflow prototype encyclopedia submodule

    Args:
        aflow_prototype_encyclopedia_path:
            Path to aflow_prototype_encyclopedia_repo

    Returns:
        A dictionary where the keys are the prototype strings, and the values are the
        shortnames found in the corresponding lines.
    """
    aflow_data_path = os.path.join(aflow_prototype_encyclopedia_path, "data")
    shortnames = {}
    for libproto in os.listdir(aflow_data_path):
        info_file = os.path.join(aflow_data_path, libproto, "info.json")
        with open(info_file) as f:
            info = json.load(f)
        shortnames[libproto] = info["title"]
    return shortnames


def get_real_to_virtual_species_map(input: Union[List[str], Atoms]) -> Dict:
    """
    Map real species to virtual species according to (alphabetized) AFLOW convention,
    e.g. for SiC return {'C':'A', 'Si':'B'}
    """
    if isinstance(input, Atoms):
        species = sorted(list(set(input.get_chemical_symbols())))
    else:
        species = input

    real_to_virtual_species_map = {}
    for i, symbol in enumerate(species):
        real_to_virtual_species_map[symbol] = chr(65 + i)

    return real_to_virtual_species_map


def solve_for_aflow_cell_params_from_primitive_ase_cell_params(
    cellpar_prim: npt.ArrayLike, prototype_label: str
) -> List[float]:
    """
    Get conventional cell parameters from primitive cell parameters. It is assumed that
    the primitive cell is related to the conventional cell as specified in
    10.1016/j.commatsci.2017.01.017. Equations obtained from Wolfram notebook in
    ``scripts/cell_param_solver.nb``

    Args:
        cellpar_prim:
            The 6 cell parameters of the primitive unit cell:
            [a, b, c, alpha, beta, gamma]
        prototype_label:
            The AFLOW prototype label of the crystal

    Returns:
        The cell parameters expected by AFLOW for the prototype label provided. The
        first parameter is always "a" and is given in the same units as
        ``cellpar_prim``, the others are fractional parameters in terms of "a", or
        angles in degrees. For example, if the ``prototype_label`` provided indicates a
        monoclinic crystal, this function will return the values of [a, b/a, c/a, beta]
    """
    bravais_lattice = get_bravais_lattice_from_prototype(prototype_label)

    assert len(cellpar_prim) == 6, "Got a number of cell parameters that is not 6"

    for length in cellpar_prim[0:3]:
        assert length > 0, "Got a negative cell size"
    for angle in cellpar_prim[3:]:
        assert 0 < angle < 180, "Got a cell angle outside of (0,180)"

    aprim = cellpar_prim[0]
    bprim = cellpar_prim[1]
    cprim = cellpar_prim[2]
    alphaprim = cellpar_prim[3]
    betaprim = cellpar_prim[4]
    gammaprim = cellpar_prim[5]

    if bravais_lattice == "aP":
        return [aprim, bprim / aprim, cprim / aprim, alphaprim, betaprim, gammaprim]
    elif bravais_lattice == "mP":
        return [aprim, bprim / aprim, cprim / aprim, betaprim]
    elif bravais_lattice == "oP":
        return [aprim, bprim / aprim, cprim / aprim]
    elif bravais_lattice == "tP" or bravais_lattice == "hP":
        return [aprim, cprim / aprim]
    elif bravais_lattice == "cP":
        return [aprim]
    elif bravais_lattice == "mC":
        cos_alphaprim = cos(radians(alphaprim))
        cos_gammaprim = cos(radians(gammaprim))
        a = aprim * sqrt(2 + 2 * cos_gammaprim)
        b = aprim * sqrt(2 - 2 * cos_gammaprim)
        c = cprim
        beta = degrees(acos(cos_alphaprim / sqrt((1 + cos_gammaprim) / 2)))
        return [a, b / a, c / a, beta]
    elif bravais_lattice == "oC":
        # the 'C' is colloquial, and can refer to either C or A-centering
        space_group_number = get_space_group_number_from_prototype(prototype_label)
        if space_group_number in C_CENTERED_ORTHORHOMBIC_GROUPS:
            cos_gammaprim = cos(radians(gammaprim))
            a = bprim * sqrt(2 + 2 * cos_gammaprim)
            b = bprim * sqrt(2 - 2 * cos_gammaprim)
            c = cprim
        elif space_group_number in A_CENTERED_ORTHORHOMBIC_GROUPS:
            cos_alphaprim = cos(radians(alphaprim))
            a = aprim
            b = bprim * sqrt(2 + 2 * cos_alphaprim)
            c = bprim * sqrt(2 - 2 * cos_alphaprim)
        else:
            raise IncorrectSpaceGroupException(
                f"Space group in prototype label {prototype_label} not found in lists "
                "of side-centered orthorhombic groups"
            )
        return [a, b / a, c / a]
    elif bravais_lattice == "oI":
        cos_alphaprim = cos(radians(alphaprim))
        cos_betaprim = cos(radians(betaprim))
        a = aprim * sqrt(2 + 2 * cos_alphaprim)
        b = aprim * sqrt(2 + 2 * cos_betaprim)
        # I guess the cosines must sum to a negative number!?
        # Will raise a ValueError: math domain error if not
        c = aprim * sqrt(-2 * (cos_alphaprim + cos_betaprim))
        return [a, b / a, c / a]
    elif bravais_lattice == "oF":
        aprimsq = aprim * aprim
        bprimsq = bprim * bprim
        cprimsq = cprim * cprim
        a = sqrt(2 * (-aprimsq + bprimsq + cprimsq))
        b = sqrt(2 * (aprimsq - bprimsq + cprimsq))
        c = sqrt(2 * (aprimsq + bprimsq - cprimsq))
        return [a, b / a, c / a]
    elif bravais_lattice == "tI":
        cos_alphaprim = cos(radians(alphaprim))
        a = aprim * sqrt(2 + 2 * cos_alphaprim)
        # I guess primitive alpha is always obtuse!? Will raise a
        # ValueError: math domain error if not
        c = 2 * aprim * sqrt(-cos_alphaprim)
        return [a, c / a]
    elif bravais_lattice == "hR":
        cos_alphaprim = cos(radians(alphaprim))
        a = aprim * sqrt(2 - 2 * cos_alphaprim)
        c = aprim * sqrt(3 + 6 * cos_alphaprim)
        return [a, c / a]
    elif bravais_lattice == "cF":
        return [aprim * sqrt(2)]
    elif bravais_lattice == "cI":
        return [aprim * 2 / sqrt(3)]


class AFLOW:
    """
    Class enabling access to the AFLOW executable

    Attributes:
        aflow_executable (str): Name of the AFLOW executable
        aflow_work_dir (str): Path to the work directory
        np (int):
            Number of processors to use, passed to the AFLOW executable using the
            ``--np=...`` argument
    """

    class AFLOWNotFoundException(Exception):
        """
        Raised when the AFLOW executable is not found
        """

    class ChangedSymmetryException(Exception):
        """
        Raised when an unexpected symmetry change is detected
        """

    class FailedToMatchException(Exception):
        """
        Raised when ``aflow --compare...`` fails to match
        """

    class FailedToSolveException(Exception):
        """
        Raised when solution algorithm fails
        """

    def __init__(
        self,
        aflow_executable: str = AFLOW_EXECUTABLE,
        aflow_work_dir: str = "",
        np: int = 4,
    ):
        """
        Args:
            aflow_executable: Sets :attr:`aflow_executable`
            aflow_work_dir: Sets :attr:`aflow_work_dir`
            np: Sets :attr:`np`
        """
        self.aflow_executable = aflow_executable
        self.np = np
        try:
            ver_str = self.get_aflow_version()
        except Exception:
            raise self.AFLOWNotFoundException(
                "Failed to run an AFLOW test command. It is likely "
                "that the AFLOW executable was not found."
            )
        # I am fine with allowing prereleases
        aflow_ver_no_prerelease = Version.parse(ver_str)
        aflow_ver_no_prerelease.replace(prerelease=None)
        if aflow_ver_no_prerelease < Version.parse(REQUIRED_AFLOW):
            raise self.AFLOWNotFoundException(
                f"Your AFLOW version {ver_str} is less "
                f"than the required {REQUIRED_AFLOW}"
            )
        if aflow_work_dir != "" and not aflow_work_dir.endswith("/"):
            self.aflow_work_dir = aflow_work_dir + "/"
        else:
            self.aflow_work_dir = aflow_work_dir

    def aflow_command(self, cmd: Union[str, List[str]], verbose=True) -> str:
        """
        Run AFLOW executable with specified arguments and return the output, possibly
        multiple times piping outputs to each other

        Args:
            cmd:
                List of arguments to pass to each AFLOW executable.
                If it's longer than 1, multiple commands will be piped to each other
            verbose: Whether to echo command to log file

        Raises:
            AFLOW.ChangedSymmetryException:
                if an ``aflow --proto=`` command complains that
                "the structure has a higher symmetry than indicated by the label"

        Returns:
            Output of the AFLOW command
        """
        if not isinstance(cmd, list):
            cmd = [cmd]

        cmd_list = [
            self.aflow_executable + " --np=" + str(self.np) + " " + cmd_inst
            for cmd_inst in cmd
        ]
        cmd_str = " | ".join(cmd_list)
        if verbose:
            logger.info(cmd_str)
        try:
            return subprocess.check_output(
                cmd_str, shell=True, stderr=subprocess.PIPE, encoding="utf-8"
            )
        except subprocess.CalledProcessError as exc:
            if "--proto=" in cmd_str and (
                "The structure has a higher symmetry than indicated by the "
                "label. The correct label and parameters for this structure are:"
            ) in str(exc.stderr):
                warn_str = (
                    "WARNING: the following command refused to write a POSCAR because "
                    f"it detected a higher symmetry: {cmd_str}. "
                    f"AFLOW error follows:\n{str(exc.stderr)}"
                )
                logger.warning(warn_str)
                raise self.ChangedSymmetryException(warn_str)
            else:
                raise exc

    def write_poscar_from_prototype(
        self,
        prototype_label: str,
        species: Optional[List[str]] = None,
        parameter_values: Optional[List[float]] = None,
        output_file: Optional[str] = None,
        verbose: bool = True,
        addtl_args: str = "",
    ) -> Optional[str]:
        """
        Run the ``aflow --proto`` command to write a POSCAR coordinate file
        corresponding to the provided AFLOW prototype designation.
        This file will have fractional coordinates.

        Args:
            prototype_label:
                An AFLOW prototype label, with or without an enumeration suffix
            species:
                List of stoichiometric species of the crystal. If this is omitted,
                the file will be written without species info
            parameter_values:
                The free parameters of the AFLOW prototype designation. If an
                enumeration suffix is not included in `prototype_label`
                and the prototype has free parameters besides `a`, this must be provided
            output_file:
                Name of the output file. If not provided,
                the output is returned as a string
            verbose: Whether to echo command to log file
            addtl_args:
                additional arguments to pass, e.g. ``--equations_only`` to get equations

        Returns:
            The output of the command or None if an `output_file` was given

        Raises:
            AFLOW.ChangedSymmetryException:
                if an ``aflow --proto=`` command complains that
                "the structure has a higher symmetry than indicated by the label"
        """
        command = " --proto=" + prototype_label
        if parameter_values:
            command += " --params=" + ",".join(
                [str(param) for param in parameter_values]
            )

        command += " " + addtl_args

        try:
            poscar_string_no_species = self.aflow_command(command, verbose=verbose)
        except self.ChangedSymmetryException as e:
            # re-raise, just indicating that this function knows about this exception
            raise e

        if species is None:
            poscar_string = poscar_string_no_species
        else:
            poscar_string = ""
            for i, line in enumerate(
                poscar_string_no_species.splitlines(keepends=True)
            ):
                poscar_string += line
                if i == 4:
                    poscar_string += " ".join(species) + "\n"
        if output_file is None:
            return poscar_string
        else:
            with open(self.aflow_work_dir + output_file, "w") as f:
                f.write(poscar_string)

    def build_atoms_from_prototype(
        self,
        prototype_label: str,
        species: List[str],
        parameter_values: Optional[List[float]] = None,
        proto_file: Optional[str] = None,
        addtl_args: str = "",
        verbose: bool = True,
    ) -> Atoms:
        """
        Build an atoms object from an AFLOW prototype designation

        Args:
            prototype_label:
                An AFLOW prototype label, with or without an enumeration suffix
            species:
                Stoichiometric species, e.g. ["Mo", "S"] corresponding to A and B
                respectively for prototype label AB2_hP6_194_c_f indicating molybdenite
            parameter_values:
                The free parameters of the AFLOW prototype designation. If an
                enumeration suffix is not included in `prototype_label`
                and the prototype has free parameters besides `a`, this must be provided
            proto_file:
                Write the POSCAR to this permanent file for debugging
                instead of a temporary file
            addtl_args:
                additional arguments to pass, e.g. ``--webpage`` to get deactivate
                higher symmetry check
            verbose:
                Print details in the log file

        Returns:
            Object representing unit cell of the material

        Raises:
            AFLOW.ChangedSymmetryException:
                if an ``aflow --proto=`` command complains that
                "the structure has a higher symmetry than indicated by the label"
        """
        try:
            poscar_string = self.write_poscar_from_prototype(
                prototype_label=prototype_label,
                species=species,
                parameter_values=parameter_values,
                addtl_args=addtl_args,
                verbose=verbose,
            )
        except self.ChangedSymmetryException as e:
            # re-raise, just indicating that this function knows about this exception
            raise e

        with (
            NamedTemporaryFile(mode="w+")
            if proto_file is None
            else open(proto_file, mode="w+")
        ) as f:
            f.write(poscar_string)
            f.seek(0)
            atoms = ase.io.read(f.name, format="vasp")
        check_number_of_atoms(atoms, prototype_label)
        atoms.wrap()
        return atoms

    def compare_materials_dir(
        self, materials_subdir: str, no_scale_volume: bool = True
    ) -> List[Dict]:
        """
        Compare a directory of materials using the aflow --compare_materials -D tool

        Args:
            materials_subdir:
                Path to the directory to compare from self.aflow_work_dir
            no_scale_volume:
                If `True`, the default behavior of allowing arbitrary scaling of
                structures before comparison is turned off

        Returns:
                Attributes of representative structures, their duplicates, and groups
                as a whole
        """
        # TODO: For efficiency, it is possible to --add_aflow_prototype_designation to
        # the representative structures. This does not help if we need duplicate
        # prototypes (for refdata), nor for library protos (as they are not ranked by
        # match like we need)
        command = " --compare_materials -D "
        command += self.aflow_work_dir + materials_subdir
        if no_scale_volume:
            command += " --no_scale_volume"
        output = self.aflow_command([command + " --screen_only --quiet --print=json"])
        res_json = json.loads(output)
        return res_json

    def get_aflow_version(self) -> str:
        """
        Run the ``aflow --version`` command to get the aflow version

        Returns:
            aflow++ executable version
        """
        command = " --version"
        output = self.aflow_command([command])
        return output.strip().split()[2]

    def compare_to_prototypes(self, input_file: str, prim: bool = True) -> List[Dict]:
        """
        Run the ``aflow --compare2prototypes`` command to compare the input structure
        to the AFLOW library of curated prototypes

        Args:
            input_file: path to the POSCAR file containing the structure to compare
            prim: whether to primitivize the structure first

        Returns:
            JSON list of dictionaries containing information about matching prototypes.
            In practice, this list should be of length zero or 1
        """
        if prim:
            command = [
                " --prim < " + self.aflow_work_dir + input_file,
                " --compare2prototypes --catalog=anrl --quiet --print=json",
            ]
        else:
            command = (
                " --compare2prototypes --catalog=anrl --quiet --print=json < "
                + self.aflow_work_dir
                + input_file
            )

        output = self.aflow_command(command)
        res_json = json.loads(output)
        return res_json

    def get_prototype_designation_from_file(
        self,
        input_file: str,
        prim: bool = True,
        force_wyckoff: bool = False,
        verbose: bool = False,
    ) -> Dict:
        """
        Run the ``aflow --prototype`` command to get the AFLOW prototype designation
            of the input structure

        Args:
            input_file: path to the POSCAR file containing the structure to analyze
            prim: whether to primitivize the structure first. Faster
            force_wyckoff:
                If the input is cif, do this to avoid re-analysis and just take the
                parameters as-is
            verbose: Whether to echo command to log file

        Returns:
            Dictionary describing the AFLOW prototype designation
            (label and parameters) of the input structure.
        """
        if prim:
            command = [
                " --prim < " + self.aflow_work_dir + input_file,
                " --prototype --print=json",
            ]
            assert not force_wyckoff, "Must specify prim=False with force_wyckoff"
        else:
            command = [
                " --prototype --print=json < " + self.aflow_work_dir + input_file
            ]

        if force_wyckoff:
            command[-1] += " --force_Wyckoff"

        output = self.aflow_command(command, verbose=verbose)
        res_json = json.loads(output)
        return res_json

    def get_prototype_designation_from_atoms(
        self, atoms: Atoms, prim: bool = True, verbose: bool = False
    ) -> Dict:
        """
        Run the ``aflow --prototype`` command to get the AFLOW prototype designation

        Args:
            atoms: atoms object to analyze
            prim: whether to primitivize the structure first
            verbose: Whether to echo command to log file

        Returns:
            Dictionary describing the AFLOW prototype designation
            (label and parameters) of the input structure.
        """
        return write_tmp_poscar_from_atoms_and_run_function(
            atoms, self.get_prototype_designation_from_file, prim=prim, verbose=verbose
        )

    def get_library_prototype_label_and_shortname_from_file(
        self, poscar_file: str, prim: bool = True, shortnames: Dict = read_shortnames()
    ) -> Tuple[Union[str, None], Union[str, None]]:
        """
        Use the aflow command line tool to determine the library prototype label for a
        structure and look up its human-readable shortname. In the case of multiple
        results, the enumeration with the smallest misfit that is in the prototypes
        list is returned. If none of the results are in the matching prototypes list,
        then the prototype with the smallest misfit is returned.

        Args:
            poscar_file:
                Path to input coordinate file
            prim: whether to primitivize the structure first
            shortnames:
                Dictionary with library prototype labels as keys and human-readable
                shortnames as values.

        Returns:
            * The library prototype label for the provided compound.
            * Shortname corresponding to this prototype
        """

        comparison_results = self.compare_to_prototypes(poscar_file, prim=prim)
        if len(comparison_results) > 1:
            # If zero results are returned it means the prototype is not in the
            # encyclopedia at all.
            # Not expecting a case where the number of results is greater than 1.
            raise RuntimeError(
                f"{comparison_results} results returned from comparison instead of "
                "zero or one as expected"
            )
        elif len(comparison_results) == 0:
            return None, None

        # Try to find the result with the smallest misfit that is in the matching
        # prototype list, otherwise return result with smallest misfit
        misfit_min_overall = 1e60
        found_overall = False
        misfit_min_inlist = 1e60
        found_inlist = False

        shortname = None
        for struct in comparison_results[0]["structures_duplicate"]:
            if struct["misfit"] < misfit_min_overall:
                misfit_min_overall = struct["misfit"]
                library_proto_overall = struct["name"]
                found_overall = True
            if struct["misfit"] < misfit_min_inlist and struct["name"] in shortnames:
                misfit_min_inlist = struct["misfit"]
                library_proto_inlist = struct["name"]
                found_inlist = True
        if found_inlist:
            matching_library_prototype_label = library_proto_inlist
            shortname = shortnames[matching_library_prototype_label]
        elif found_overall:
            matching_library_prototype_label = library_proto_overall
        else:
            matching_library_prototype_label = None

        logger.info(
            "Detected encyclopedia entry "
            f"{matching_library_prototype_label}: {shortname}"
        )
        return matching_library_prototype_label, shortname

    def get_library_prototype_label_and_shortname_from_atoms(
        self, atoms: Atoms, prim: bool = True, shortnames: Dict = read_shortnames()
    ) -> Tuple[Union[str, None], Union[str, None]]:
        """
        Use the aflow command line tool to determine the library prototype label for a
        structure and look up its human-readable shortname. In the case of multiple
        results, the enumeration with the smallest misfit that is in the prototypes
        list is returned. If none of the results are in the matching prototypes list,
        then the prototype with the smallest misfit is returned.

        Args:
            atoms:
                Atoms object to compare
            prim: whether to primitivize the structure first
            shortnames:
                Dictionary with library prototype labels as keys and human-readable
                shortnames as values.

        Returns:
            * The library prototype label for the provided compound.
            * Shortname corresponding to this prototype
        """
        return write_tmp_poscar_from_atoms_and_run_function(
            atoms,
            self.get_library_prototype_label_and_shortname_from_file,
            prim=prim,
            shortnames=shortnames,
        )

    def _compare_poscars(self, poscar1: PathLike, poscar2: PathLike) -> Dict:
        return json.loads(
            self.aflow_command(
                [
                    f" --print=JSON --compare_materials={poscar1},{poscar2} "
                    "--screen_only --no_scale_volume --optimize_match --quiet"
                ],
                verbose=False,
            )
        )

    def _compare_Atoms(
        self,
        atoms1: Atoms,
        atoms2: Atoms,
        sort_atoms1: bool = True,
        sort_atoms2: bool = True,
    ) -> Dict:
        with NamedTemporaryFile() as f1, NamedTemporaryFile() as f2:
            atoms1.write(f1.name, "vasp", sort=sort_atoms1)
            atoms2.write(f2.name, "vasp", sort=sort_atoms2)
            f1.seek(0)
            f2.seek(0)
            compare = self._compare_poscars(f1.name, f2.name)
        return compare

    def get_basistransformation_rotation_originshift_atom_map_from_atoms(
        self,
        atoms1: Atoms,
        atoms2: Atoms,
        sort_atoms1: bool = True,
        sort_atoms2: bool = True,
    ) -> Tuple[
        Optional[npt.ArrayLike],
        Optional[npt.ArrayLike],
        Optional[npt.ArrayLike],
        Optional[List[int]],
    ]:
        """
        Get operations to transform atoms2 to atoms1

        Args:
            sort_atoms1:
                Whether to sort atoms1 before comparing. If species are not
                alphabetized, this is REQUIRED. However, the `atom_map` returned will
                be w.r.t the sorted order, use with care!
            sort_atoms2: Whether to sort atoms2 before comparing.

        Returns:
            * basis transformation
            * rotation
            * origin shift
            * atom_map (atom_map[index_in_structure_1] = index_in_structure_2)

        Raises:
            AFLOW.FailedToMatchException: if AFLOW fails to match the crystals
        """
        comparison_result = self._compare_Atoms(
            atoms1, atoms2, sort_atoms1, sort_atoms2
        )
        if (
            "structures_duplicate" in comparison_result[0]
            and comparison_result[0]["structures_duplicate"] != []
        ):
            return (
                np.asarray(
                    comparison_result[0]["structures_duplicate"][0][
                        "basis_transformation"
                    ]
                ),
                np.asarray(comparison_result[0]["structures_duplicate"][0]["rotation"]),
                np.asarray(
                    comparison_result[0]["structures_duplicate"][0]["origin_shift"]
                ),
                comparison_result[0]["structures_duplicate"][0]["atom_map"],
            )
        else:
            msg = "AFLOW was unable to match the provided crystals"
            logger.info(msg)
            raise self.FailedToMatchException(msg)

    def get_param_names_from_prototype(self, prototype_label: str) -> List[str]:
        """
        Get the parameter names
        """
        symbol_string = self.write_poscar_from_prototype(
            prototype_label=prototype_label, addtl_args="--parameter_symbols_only"
        )
        return symbol_string.strip().split(",")

    def get_equation_sets_from_prototype(
        self, prototype_label: str
    ) -> List[EquivalentEqnSet]:
        """
        Get the symbolic equations for the fractional positions in the unit cell of an
        AFLOW prototype

        Args:
            prototype_label:
                An AFLOW prototype label, without an enumeration suffix, without
                specified atomic species

        Returns:
            List of EquivalentEqnSet objects

            Each EquivalentEqnSet contains:
                - species: The species of the atoms in this set.
                - wyckoff_letter: The Wyckoff letter corresponding to this set.
                - param_names: The names of the free parameters associated with this
                  Wyckoff position.
                - coeff_matrix_list: A list of 3 x n matrices of coefficients for the
                  free parameters.
                - const_terms_list: A list of 3 x 1 columns of constant terms in the
                  coordinates.
        """
        equation_poscar = self.write_poscar_from_prototype(
            prototype_label,
            addtl_args="--equations_only",
        )

        # get a string with one character per Wyckoff position
        # (with possible repeated letters for positions with free params)
        wyckoff_lists = get_wyckoff_lists_from_prototype(prototype_label)
        wyckoff_joined_list = "".join(wyckoff_lists)

        coord_lines = equation_poscar.splitlines()
        coord_iter = iter(coord_lines)
        reading_coord = False
        while not reading_coord:
            line = next(coord_iter)
            if line.startswith("BEGIN EQUATIONS WITH WYCKOFF"):
                reading_coord = True

        space_group_number = get_space_group_number_from_prototype(prototype_label)

        equation_sets = []

        for wyckoff_letter in wyckoff_joined_list:
            species = None  # have not seen a line yet, so don't know what species it is
            param_names = None  # same as above.
            coeff_matrix_list = []
            const_terms_list = []
            # the next n positions should be equivalent
            # corresponding to this Wyckoff position
            for _ in range(
                get_primitive_wyckoff_multiplicity(space_group_number, wyckoff_letter)
            ):
                line_split = next(coord_iter).split()
                if species is None:
                    species = line_split[4]
                elif line_split[4] != species:
                    raise InconsistentWyckoffException(
                        "Encountered different species within what I thought should be "
                        f"the lines corresponding to Wyckoff position {wyckoff_letter}"
                        f"\nEquations obtained from prototype label {prototype_label}:"
                        f"\n{equation_poscar}"
                    )
                # first, get the free parameters of this line
                curr_line_free_params = set()  # sympy.Symbol
                coordinate_expr_list = []
                for expression_string in line_split[1:4]:
                    coordinate_expr = parse_expr(expression_string)
                    curr_line_free_params.update(coordinate_expr.free_symbols)
                    coordinate_expr_list.append(coordinate_expr)

                # They should all have the same number, i.e. x2, y2, z2 or x14, z14,
                # so we can just string sort them
                curr_line_free_params = list(curr_line_free_params)
                curr_line_free_params.sort(key=lambda param: str(param))

                # Each line within a Wyckoff position should
                # have the same set of free parameters
                if param_names is None:
                    param_names = curr_line_free_params
                elif param_names != curr_line_free_params:
                    raise InconsistentWyckoffException(
                        "Encountered different free params within what I thought "
                        "should be the lines corresponding to Wyckoff position "
                        f"{wyckoff_letter}\nEquations obtained from prototype label "
                        f"{prototype_label}:\n{equation_poscar}"
                    )

                # Transform to matrices and vectors
                a, b = linear_eq_to_matrix(coordinate_expr_list, param_names)

                assert a.shape == (3, len(param_names))
                assert b.shape == (3, 1)

                coeff_matrix_list.append(matrix2numpy(a, dtype=np.float64))
                const_terms_list.append(matrix2numpy(-b, dtype=np.float64))

            # Done looping over this set of equivalent positions
            equation_sets.append(
                EquivalentEqnSet(
                    species=species,
                    wyckoff_letter=wyckoff_letter,
                    param_names=[str(param_name) for param_name in param_names],
                    coeff_matrix_list=coeff_matrix_list,
                    const_terms_list=const_terms_list,
                )
            )

        # do some checks
        equation_sets_iter = iter(equation_sets)
        species = None
        for species_wyckoff_list in wyckoff_lists:
            species_must_change = True
            for _ in species_wyckoff_list:
                equation_set = next(equation_sets_iter)
                if species_must_change:
                    if equation_set.species == species:
                        raise InconsistentWyckoffException(
                            "The species in the equations obtained below are "
                            "inconsistent with the number and multiplicity of Wyckoff "
                            f"positions in prototype label {prototype_label}\n"
                            f"{equation_poscar}"
                        )
                species = equation_set.species
                species_must_change = False

        return equation_sets

    def solve_for_params_of_known_prototype(
        self,
        atoms: Atoms,
        prototype_label: str,
        max_resid: Optional[float] = None,
        cell_rtol: float = 0.01,
        rot_rtol: float = 0.01,
        rot_atol: float = 0.01,
    ) -> Tuple[List[float], Optional[str]]:
        """
        Given an Atoms object that is a primitive cell of its Bravais lattice as
        defined in doi.org/10.1016/j.commatsci.2017.01.017, and its presumed prototype
        label, solves for the free parameters of the prototype label. Raises an error
        if the solution fails (likely indicating that the Atoms object provided does
        not conform to the provided prototype label.) The Atoms object may be rotated,
        translated, and permuted, but the identity of the lattice vectors must be
        unchanged w.r.t. the crystallographic prototype. In other words, there must
        exist a permutation and translation of the fractional coordinates that enables
        them to match the equations defined by the prototype label.

        Args:
            atoms: The Atoms object to analyze
            prototype_label:
                The assumed AFLOW prototype label. `atoms` must be a primitive cell
                (with lattice vectors defined in
                doi.org/10.1016/j.commatsci.2017.01.017) conforming to the symmetry
                defined by this prototype label. Rigid body rotations, translations,
                and permutations of `atoms` relative to the AFLOW setting are allowed,
                but the identity of the lattice vectors must be unchanged.
            max_resid:
                Maximum residual allowed when attempting to match the fractional
                positions of the atoms to the crystallographic equations
                If not provided, this is automatically set to 0.01*(minimum NN distance)
            cell_rtol:
                Relative tolerance on cell lengths and angles
                Justification for default value: AFLOW uses 0.01*(minimum NN distance)
                as default tolerance.
            rot_rtol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)
            rot_atol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)

        Returns:
            * List of free parameters that will regenerate `atoms` (up to permutations,
              rotations, and translations) when paired with `prototype_label`
            * Library prototype label from the AFLOW prototype encyclopedia, if any
            * Title of library prototype from the AFLOW prototype encyclopedia, if any

        Raises:
            AFLOW.ChangedSymmetryException:
                if the symmetry of the atoms object is different from `prototype_label`
            AFLOW.FailedToMatchException:
                if AFLOW fails to match the re-generated crystal to the input crystal

        """
        # If max_resid not provided, determine it from neighborlist
        if max_resid is None:
            nl_len = 0
            cov_mult = 1
            while nl_len == 0:
                logger.info(
                    "Attempting to find NN distance by searching "
                    f"within covalent radii times {cov_mult}"
                )
                nl = neighbor_list("d", atoms, natural_cutoffs(atoms, mult=cov_mult))
                nl_len = nl.size
                cov_mult += 1
            # set the maximum error to 1% of NN distance to follow AFLOW convention
            # rescale by cube root of cell volume to get rough conversion from
            # cartesian to fractional
            max_resid = nl.min() * 0.01 * atoms.get_volume() ** (-1 / 3)
            logger.info(
                "Automatically set max fractional residual for solving position "
                f"equations to {max_resid}"
            )

        # solve for cell parameters
        cell_params = solve_for_aflow_cell_params_from_primitive_ase_cell_params(
            atoms.cell.cellpar(), prototype_label
        )
        species = sorted(list(set(atoms.get_chemical_symbols())))

        # First, redetect the prototype label. We can't use this as-is because it may
        # be rotated by an operation that's within the normalizer but not
        # in the space group itself
        detected_prototype_designation = self.get_prototype_designation_from_atoms(
            atoms
        )

        prototype_label_detected = detected_prototype_designation[
            "aflow_prototype_label"
        ]

        library_prototype_label, short_name = (
            self.get_library_prototype_label_and_shortname_from_atoms(atoms)
        )

        if not prototype_labels_are_equivalent(
            prototype_label, prototype_label_detected
        ):
            msg = (
                f"Redetected prototype label {prototype_label_detected} does not match "
                f"nominal {prototype_label}."
            )
            logger.info(msg)
            raise self.ChangedSymmetryException(msg)

        # rebuild the atoms
        try:
            atoms_rebuilt = self.build_atoms_from_prototype(
                prototype_label=prototype_label_detected,
                species=species,
                parameter_values=detected_prototype_designation[
                    "aflow_prototype_params_values"
                ],
                addtl_args="--webpage",
            )
        except self.ChangedSymmetryException as e:
            # re-raise, just indicating that this function knows about this exception
            raise e

        # We want the negative of the origin shift from ``atoms_rebuilt`` to ``atoms``,
        # because the origin shift is the last operation to happen, so it will be in
        # the ``atoms`` frame.

        # This function gets the transformation from its second argument to its first.
        # The origin shift is Cartesian if the POSCARs are Cartesian, which they are
        # when made from Atoms

        # Sort atoms, but do not sort atoms_rebuilt
        try:
            _, _, neg_initial_origin_shift_cart, atom_map = (
                self.get_basistransformation_rotation_originshift_atom_map_from_atoms(
                    atoms, atoms_rebuilt, sort_atoms1=True, sort_atoms2=False
                )
            )
        except self.FailedToMatchException:
            # Re-raise with a more informative error message
            msg = (
                "Execution cannot continue because AFLOW failed to match the crystal "
                "with its representation re-generated from the detected prototype "
                "designation.\nRarely, this can happen if the structure is on the edge "
                "of a symmetry increase (e.g. a BCT structure with c/a very close to 1)"
            )
            logger.error(msg)
            raise self.FailedToMatchException(msg)

        # Transpose the change of basis equation for row vectors
        initial_origin_shift_frac = (-neg_initial_origin_shift_cart) @ np.linalg.inv(
            atoms.cell
        )

        logger.info(
            "Initial shift (to SOME standard origin, not necessarily the desired one): "
            f"{-neg_initial_origin_shift_cart} (Cartesian), "
            f"{initial_origin_shift_frac} (fractional)"
        )

        position_set_list = get_equivalent_atom_sets_from_prototype_and_atom_map(
            atoms, prototype_label, atom_map, sort_atoms=True
        )

        # get equation sets
        equation_set_list = self.get_equation_sets_from_prototype(prototype_label)

        if len(position_set_list) != len(equation_set_list):
            raise InconsistentWyckoffException(
                "Number of equivalent positions detected in Atoms object "
                "did not match the number of equivalent equations given"
            )

        space_group_number = get_space_group_number_from_prototype(prototype_label)
        for prim_shift in get_possible_primitive_shifts(space_group_number):
            logger.info(
                "Additionally shifting atoms by "
                f"internal fractional translation {prim_shift}"
            )

            free_params_dict = {}
            position_set_matched_list = [False] * len(position_set_list)

            for equation_set in equation_set_list:
                # Because both equations and positions are sorted by species and
                # wyckoff letter, this should be pretty efficient
                matched_this_equation_set = False
                for i, position_set in enumerate(position_set_list):
                    if position_set_matched_list[i]:
                        continue
                    if (
                        position_set.species != equation_set.species
                    ):  # These are virtual species
                        continue
                    if not are_in_same_wyckoff_set(
                        equation_set.wyckoff_letter,
                        position_set.wyckoff_letter,
                        space_group_number,
                    ):
                        continue
                    for coeff_matrix, const_terms in zip(
                        equation_set.coeff_matrix_list, equation_set.const_terms_list
                    ):
                        for frac_position in position_set.frac_position_list:
                            # Here we use column coordinate vectors
                            frac_position_shifted = (
                                frac_position
                                + np.asarray(prim_shift).reshape(3, 1)
                                + initial_origin_shift_frac.reshape(3, 1)
                            ) % 1
                            possible_shifts = (-1, 0, 1)
                            # explore all possible shifts around zero
                            # to bring back in cell.
                            # TODO: if this is too slow (27 possibilities), write an
                            # algorithm to determine which shifts are possible
                            for shift_list in [
                                (x, y, z)
                                for x in possible_shifts
                                for y in possible_shifts
                                for z in possible_shifts
                            ]:
                                shift_array = np.asarray(shift_list).reshape(3, 1)
                                candidate_internal_param_values, resid, _, _ = (
                                    np.linalg.lstsq(
                                        coeff_matrix,
                                        frac_position_shifted
                                        - const_terms
                                        - shift_array,
                                    )
                                )
                                if len(resid) == 0 or np.max(resid) < max_resid:
                                    assert len(candidate_internal_param_values) == len(
                                        equation_set.param_names
                                    )
                                    for param_name, param_value in zip(
                                        equation_set.param_names,
                                        candidate_internal_param_values,
                                    ):
                                        assert param_name not in free_params_dict
                                        free_params_dict[param_name] = (
                                            param_value[0] % 1
                                        )  # wrap to [0,1)
                                    # should only need one to match to check off this
                                    # Wyckoff position
                                    position_set_matched_list[i] = True
                                    matched_this_equation_set = True
                                    break
                                # end loop over shifts
                            if matched_this_equation_set:
                                break
                            # end loop over positions within a position set
                        if matched_this_equation_set:
                            break
                        # end loop over equations within an equation set
                    if matched_this_equation_set:
                        break
                    # end loop over position sets
                # end loop over equation sets

            if all(position_set_matched_list):
                candidate_prototype_param_values = cell_params + [
                    free_params_dict[key]
                    for key in sorted(
                        free_params_dict.keys(), key=internal_parameter_sort_key
                    )
                ]
                # The internal shift may have taken us to an internal parameter
                # solution that represents a rotation, so we need to check
                if self.confirm_unrotated_prototype_designation(
                    reference_atoms=atoms,
                    species=species,
                    prototype_label=prototype_label,
                    parameter_values=candidate_prototype_param_values,
                    cell_rtol=cell_rtol,
                    rot_rtol=rot_rtol,
                    rot_atol=rot_atol,
                ):
                    logger.info(
                        f"Found set of parameters for prototype {prototype_label} "
                        "that is unrotated"
                    )
                    return (
                        candidate_prototype_param_values,
                        library_prototype_label,
                        short_name,
                    )
                else:
                    logger.info(
                        f"Found set of parameters for prototype {prototype_label}, "
                        "but it was rotated relative to the original cell"
                    )
            else:
                logger.info(
                    f"Failed to solve equations for prototype {prototype_label} "
                    "on this shift attempt"
                )

        msg = (
            f"Failed to solve equations for prototype {prototype_label} "
            "on any shift attempt"
        )
        logger.info(msg)
        raise self.FailedToSolveException(msg)

    def confirm_atoms_unrotated_when_cells_aligned(
        self,
        test_atoms: Atoms,
        ref_atoms: Atoms,
        sgnum: Union[int, str],
        cell_rtol: float = 0.01,
        rot_rtol: float = 0.01,
        rot_atol: float = 0.01,
    ) -> bool:
        """
        Check whether `test_atoms` and `reference_atoms` are unrotated as follows:
        When the cells are in :meth:`ase.cell.Cell.standard_form`, the cells are
        identical. When both crystals are rotated to standard form (rotating the cell
        and keeping the fractional coordinates unchanged), the rotation part of the
        mapping the two crystals to each other found by AFLOW is in the point group of
        the reference crystal (using the generated crystal would give the same result).
        In other words, the crystals are identically oriented (but possibly translated)
        in reference to their lattice vectors, which in turn must be identical up to a
        rotation in reference to some Cartesian coordinate system.
        The crystals must be primitive cells as defined in
        https://doi.org/10.1016/j.commatsci.2017.01.017.

        Args:
            test_atoms:
                Primitive cell of a crystal
            ref_atoms:
                Primitive cell of a crystal
            sgnum:
                Space group number
            cell_rtol:
                Parameter to pass to :func:`numpy.allclose` for comparing cell params.
                Justification for default value: AFLOW uses 0.01*(minimum NN distance)
                as default tolerance.
            rot_rtol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)
            rot_atol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)
        """
        if not np.allclose(
            ref_atoms.cell.cellpar(), test_atoms.cell.cellpar(), rtol=cell_rtol
        ):
            logger.info(
                "Cell lengths and angles do not match.\n"
                f"Original: {ref_atoms.cell.cellpar()}\n"
                f"Regenerated: {test_atoms.cell.cellpar()}"
            )
            return False
        else:
            cell_lengths_and_angles = ref_atoms.cell.cellpar()

        test_atoms_copy = test_atoms.copy()
        ref_atoms_copy = ref_atoms.copy()

        test_atoms_copy.set_cell(
            Cell.fromcellpar(cell_lengths_and_angles), scale_atoms=True
        )
        ref_atoms_copy.set_cell(
            Cell.fromcellpar(cell_lengths_and_angles), scale_atoms=True
        )

        # the rotation below is Cartesian.
        try:
            _, cart_rot, _, _ = (
                self.get_basistransformation_rotation_originshift_atom_map_from_atoms(
                    test_atoms_copy, ref_atoms_copy
                )
            )
        except self.FailedToMatchException:
            logger.info("AFLOW failed to match the recreated crystal to reference")
            return False

        return cartesian_rotation_is_in_point_group(
            cart_rot=cart_rot,
            sgnum=sgnum,
            cell=test_atoms_copy.cell,
            rtol=rot_rtol,
            atol=rot_atol,
        )

    def confirm_unrotated_prototype_designation(
        self,
        reference_atoms: Atoms,
        species: List[str],
        prototype_label: str,
        parameter_values: List[float],
        cell_rtol: float = 0.01,
        rot_rtol: float = 0.01,
        rot_atol: float = 0.01,
    ) -> bool:
        """
        Check whether the provided prototype designation recreates ``reference_atoms``
        as follows: When the cells are in :meth:`ase.cell.Cell.standard_form`, the
        cells are identical. When both crystals are rotated to standard form (rotating
        the cell and keeping the fractional coordinates unchanged), the rotation part
        of the mapping the two crystals to each other found by AFLOW is in the point
        group of the reference crystal (using the generated crystal would give the same
        result). In other words, the crystals are identically oriented (but possibly
        translated) in reference to their lattice vectors, which in turn must be
        identical up to a rotation in reference to some Cartesian coordinate system.

        Args:
            species:
                Stoichiometric species, e.g. ["Mo", "S"] corresponding to A and B
                respectively for prototype label AB2_hP6_194_c_f indicating molybdenite
            prototype_label:
                An AFLOW prototype label, without an enumeration suffix,
                without specified atomic species
            parameter_values:
                The free parameters of the AFLOW prototype designation
            cell_rtol:
                Parameter to pass to :func:`numpy.allclose` for comparing cell params
                Justification for default value: AFLOW uses 0.01*(minimum NN distance)
                as default tolerance.
            rot_rtol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)
            rot_atol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)

        Returns:
            Whether or not the crystals match
        """
        test_atoms = self.build_atoms_from_prototype(
            prototype_label=prototype_label,
            species=species,
            parameter_values=parameter_values,
            addtl_args="--webpage",
        )

        return self.confirm_atoms_unrotated_when_cells_aligned(
            test_atoms=test_atoms,
            ref_atoms=reference_atoms,
            sgnum=get_space_group_number_from_prototype(prototype_label),
            cell_rtol=cell_rtol,
            rot_rtol=rot_rtol,
            rot_atol=rot_atol,
        )



#!/usr/bin/python

import os

import kim_edn
import numpy as np
import numpy.typing as npt
from ase.atoms import Atoms
from ase.build import bulk
from ase.calculators.lj import LennardJones

from kim_tools import (
    KIMTestDriver,
    SingleCrystalTestDriver,
    detect_unique_crystal_structures,
    get_deduplicated_property_instances,
)
from kim_tools.test_driver.core import _get_optional_source_value


class TestTestDriver(KIMTestDriver):
    def _calculate(self, property_name, species):
        """
        example calculate method

        Args:
            property_name: for testing ability to find properties at different paths.
            !!! AN ACTUAL TEST DRIVER SHOULD NOT HAVE AN ARGUMENT SUCH AS THIS !!!
        """
        atoms = Atoms([species], [[0, 0, 0]])
        self._add_property_instance(property_name, "This is an example disclaimer.")
        self._add_key_to_current_property_instance(
            "species", atoms.get_chemical_symbols()[0]
        )
        self._add_key_to_current_property_instance(
            "mass", atoms.get_masses()[0], "amu", {"source-std-uncert-value": 1}
        )


class TestSingleCrystalTestDriver(SingleCrystalTestDriver):
    def _calculate(self, deform_matrix: npt.NDArray = np.eye(3), **kwargs):
        """
        strain the crystal and write a crystal-structure-npt.
        For testing various crystal detection things
        """
        atoms = self._get_atoms()
        original_cell = atoms.cell
        new_cell = original_cell @ deform_matrix
        atoms.set_cell(new_cell, scale_atoms=True)
        self._update_nominal_parameter_values(atoms)
        self._add_property_instance_and_common_crystal_genome_keys(
            "crystal-structure-npt"
        )


def test_kimtest(monkeypatch):
    test = TestTestDriver(LennardJones())
    testing_property_names = [
        "atomic-mass",  # already in kim-properties
        "atomic-mass0",  # found in $PWD/local-props
        "atomic-mass0",  # check that repeat works fine
        # check that full id works as well, found in $PWD/local-props
        "tag:brunnels@noreply.openkim.org,2016-05-11:property/atomic-mass1",
        "atomic-mass2",  # found in $PWD/local-props/atomic-mass2
        "atomic-mass3",  # found in $PWD/mock-test-drivers-dir/mock-td/local_props,
        # tested using the monkeypatch below
    ]

    monkeypatch.setenv(
        "KIM_PROPERTY_PATH",
        os.path.join(os.getcwd(), "mock-test-drivers-dir/*/local-props")
        + ":"
        + os.path.join(os.getcwd(), "mock-test-drivers-dir/*/local_props"),
    )

    for prop_name in testing_property_names:
        test(property_name=prop_name, species="Ar")

    assert len(test.property_instances) == 6
    test.write_property_instances_to_file()


def test_detect_unique_crystal_structures():
    reference_structure = kim_edn.load("structures/OSi.edn")
    test_structure = kim_edn.load("structures/OSi_twin.edn")
    assert (
        len(
            detect_unique_crystal_structures(
                [
                    reference_structure,
                    reference_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                ],
                allow_rotation=True,
            )
        )
        == 1
    )
    assert (
        len(
            detect_unique_crystal_structures(
                [
                    reference_structure,
                    reference_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                ],
                allow_rotation=False,
            )
        )
        == 2
    )


def test_get_deduplicated_property_instances():
    property_instances = kim_edn.load("structures/results.edn")
    fully_deduplicated = get_deduplicated_property_instances(property_instances)
    assert len(fully_deduplicated) == 6
    inst_with_1_source = 0
    inst_with_2_source = 0
    for property_instance in fully_deduplicated:
        n_inst = len(
            property_instance["crystal-genome-source-structure-id"]["source-value"][0]
        )
        if n_inst == 1:
            inst_with_1_source += 1
        elif n_inst == 2:
            inst_with_2_source += 1
        else:
            assert False
    assert inst_with_1_source == 3
    assert inst_with_2_source == 3
    partially_deduplicated = get_deduplicated_property_instances(
        property_instances, ["mass-density-crystal-npt"]
    )
    assert len(partially_deduplicated) == 8
    inst_with_1_source = 0
    inst_with_2_source = 0
    for property_instance in partially_deduplicated:
        n_inst = len(
            property_instance["crystal-genome-source-structure-id"]["source-value"][0]
        )
        if n_inst == 1:
            inst_with_1_source += 1
        elif n_inst == 2:
            inst_with_2_source += 1
        else:
            assert False
    assert inst_with_1_source == 7
    assert inst_with_2_source == 1


def test_structure_detection():
    test = TestSingleCrystalTestDriver(LennardJones())
    atoms = bulk("Mg")
    hcp_prototype = "A_hP2_194_c"
    hcp_library_prototype = "A_hP2_194_c-001"
    hcp_shortname = ["Hexagonal Close Packed (Mg, $A3$, hcp) Structure"]
    stretch = np.diag([1, 1, 10])
    for deform, prototype, library_prototype, shortname in zip(
        [np.eye(3), stretch],
        [hcp_prototype, hcp_prototype],
        [hcp_library_prototype, None],
        [hcp_shortname, None],
    ):
        property_instance = test(atoms, deform_matrix=deform)[0]
        assert property_instance["prototype-label"]["source-value"] == prototype
        assert (
            _get_optional_source_value(property_instance, "library-prototype-label")
            == library_prototype
        )
        assert _get_optional_source_value(property_instance, "short-name") == shortname


if __name__ == "__main__":
    test_structure_detection()
